import polars as pl
import pytest

from maplib import (
    Model,
    Template,
    IRI,
    Prefix,
    Triple,
    Variable,
    Parameter,
    Literal,
    xsd,
    RDFType,
)
from polars.testing import assert_frame_equal

def test_write_empty_model():
    m = Model()
    s = m.writes()

def test_create_model_from_empty_polars_df():
    doc = """
    @prefix ex:<http://example.net/ns#>.
    ex:ExampleTemplate [?MyValue] :: {
        ottr:Triple(ex:myObject, ex:hasValue, ?MyValue)
    } .
    """

    df = pl.DataFrame({"MyValue": []})
    model = Model()
    model.add_template(doc)
    model.map("http://example.net/ns#ExampleTemplate", df)


def test_add_template_instead_of_constructor_df():
    doc = """
    @prefix ex:<http://example.net/ns#>.
    ex:ExampleTemplate [?MyValue] :: {
    ottr:Triple(ex:myObject, ex:hasValue, ?MyValue)
    } .
    """

    df = pl.DataFrame({"MyValue": []})
    model = Model()
    model.add_template(doc)
    model.map("http://example.net/ns#ExampleTemplate", df)


@pytest.mark.parametrize("streaming", [True, False])
def test_create_model_with_optional_value_missing_df(streaming):
    doc = """
    @prefix ex:<http://example.net/ns#>.
    ex:ExampleTemplate [?MyValue, ??MyOtherValue] :: {
    ottr:Triple(ex:myObject, ex:hasValue, ?MyValue),
    ottr:Triple(ex:myObject, ex:hasOtherValue, ?MyOtherValue)
    } .
    """

    df = pl.DataFrame({"MyValue": ["A"]})
    model = Model()
    model.add_template(doc)
    model.map("http://example.net/ns#ExampleTemplate", df)
    qres = model.query(
        """
    PREFIX ex:<http://example.net/ns#>
    
    SELECT ?A WHERE {
    ?obj1 ex:hasValue ?A
    } 
    """,
        streaming=streaming,
    )
    expected_df = pl.DataFrame(
        {
            "A": ["A"],
        }
    )
    assert_frame_equal(qres, expected_df)

    qres = model.query(
        """
        PREFIX ex:<http://example.net/ns#>
        
        SELECT ?A WHERE {
        ?obj1 ex:hasOtherValue ?A
        } 
        """,
        streaming=streaming,
    )
    assert qres.height == 0


@pytest.mark.parametrize("streaming", [True, False])
def test_create_programmatic_model_with_optional_value_missing_df(streaming):
    df = pl.DataFrame({"MyValue": ["A"]})
    model = Model()
    ex = Prefix("http://example.net/ns#")
    my_value = Variable("MyValue")
    my_other_value = Variable("MyOtherValue")
    my_object = ex.suf("MyObject")
    template = Template(
        ex.suf("ExampleTemplate"),
        [Parameter(my_value), Parameter(my_other_value, optional=True)],
        [
            Triple(my_object, ex.suf("hasValue"), my_value),
            Triple(my_object, ex.suf("hasOtherValue"), my_other_value),
        ],
    )
    model.map(template, df)
    qres = model.query(
        """
        PREFIX ex:<http://example.net/ns#>
        
        SELECT ?A WHERE {
        ?obj1 ex:hasValue ?A
        } 
        """,
        streaming=streaming,
    )
    expected_df = pl.DataFrame(
        {
            "A": ["A"],
        }
    )
    assert_frame_equal(qres, expected_df)

    qres = model.query(
        """
        PREFIX ex:<http://example.net/ns#>
        
        SELECT ?A WHERE {
        ?obj1 ex:hasOtherValue ?A
        } 
        """,
        streaming=streaming,
    )
    assert qres.height == 0


@pytest.mark.parametrize("streaming", [True, False])
def test_create_programmatic_model_with_default(streaming):
    df = pl.DataFrame({"MyValue": ["A"]})
    mapping = Model()
    ex = Prefix("http://example.net/ns#")
    my_value = Variable("MyValue")
    my_other_value = Variable("MyOtherValue")
    yet_another_value = Variable("YetAnotherValue")
    my_object = ex.suf("MyObject")
    template = Template(
        ex.suf("ExampleTemplate"),
        [
            Parameter(my_value),
            Parameter(my_other_value, default_value=Literal("123")),
            Parameter(
                yet_another_value, default_value=Literal("123", data_type=xsd.integer)
            ),
        ],
        [
            Triple(my_object, ex.suf("hasValue"), my_value),
            Triple(my_object, ex.suf("hasOtherValue"), my_other_value),
            Triple(my_object, ex.suf("hasYetAnotherValue"), yet_another_value),
        ],
    )
    mapping.map(template, df)
    qres = mapping.query(
        """
        PREFIX ex:<http://example.net/ns#>
        
        SELECT ?A ?B WHERE {
        ?obj1 ex:hasOtherValue ?A .
        ?obj1 ex:hasYetAnotherValue ?B .
        } 
        """,
        streaming=streaming,
    )
    expected_df = pl.DataFrame({"A": ["123"], "B": [123]})
    assert_frame_equal(qres, expected_df)


@pytest.mark.parametrize("streaming", [True, False])
def test_create_programmatic_mapping_with_nested_default(streaming):
    
    df = pl.DataFrame({"MyValue": ["A"]})
    mapping = Model()
    ex = Prefix("http://example.net/ns#")
    my_value = Variable("MyValue")
    my_other_value = Variable("MyOtherValue")
    yet_another_value = Variable("YetAnotherValue")
    my_object = ex.suf("MyObject")
    template = Template(
        ex.suf("ExampleTemplate"),
        [
            Parameter(my_value),
            Parameter(my_other_value, default_value=Literal("123")),
            Parameter(
                yet_another_value, default_value=Literal("123", data_type=xsd.integer)
            ),
        ],
        [
            Triple(my_object, ex.suf("hasValue"), my_value),
            Triple(my_object, ex.suf("hasOtherValue"), my_other_value),
            Triple(my_object, ex.suf("hasYetAnotherValue"), yet_another_value),
        ],
    )
    template2 = Template(
        ex.suf("ExampleTemplate2"),
        [Parameter(my_value)],
        [template.instance([my_value, None, None])],
    )
    mapping.add_template(template)
    mapping.map(template2, df)
    qres = mapping.query(
        """
        PREFIX ex:<http://example.net/ns#>
        
        SELECT ?A ?B WHERE {
        ?obj1 ex:hasOtherValue ?A .
        ?obj1 ex:hasYetAnotherValue ?B .
        } 
        """,
        streaming=streaming,
    )
    expected_df = pl.DataFrame({"A": ["123"], "B": [123]})
    assert_frame_equal(qres, expected_df)


@pytest.mark.parametrize("streaming", [True, False])
def test_create_programmatic_mapping_with_nested_default_and_missing_column(streaming):
    
    df = pl.DataFrame({"MyValue": ["A"]})
    mapping = Model()
    ex = Prefix("http://example.net/ns#")
    my_value = Variable("MyValue")
    my_other_value = Variable("MyOtherValue")
    yet_another_value = Variable("YetAnotherValue")
    my_object = ex.suf("MyObject")
    template = Template(
        ex.suf("ExampleTemplate"),
        [
            Parameter(my_value),
            Parameter(my_other_value, default_value=Literal("123")),
            Parameter(
                yet_another_value, default_value=Literal("123", data_type=xsd.integer)
            ),
        ],
        [
            Triple(my_object, ex.suf("hasValue"), my_value),
            Triple(my_object, ex.suf("hasOtherValue"), my_other_value),
            Triple(my_object, ex.suf("hasYetAnotherValue"), yet_another_value),
        ],
    )
    template2 = Template(
        ex.suf("ExampleTemplate2"),
        [
            Parameter(my_value),
            Parameter(my_other_value, optional=True),
            Parameter(yet_another_value, optional=True),
        ],
        [template.instance([my_value, my_other_value, yet_another_value])],
    )
    mapping.add_template(template)
    mapping.map(template2, df)
    qres = mapping.query(
        """
        PREFIX ex:<http://example.net/ns#>
        
        SELECT ?A ?B WHERE {
        ?obj1 ex:hasOtherValue ?A .
        ?obj1 ex:hasYetAnotherValue ?B .
        } 
        """,
        streaming=streaming,
    )
    expected_df = pl.DataFrame({"A": ["123"], "B": [123]})
    assert_frame_equal(qres, expected_df)


@pytest.mark.parametrize("streaming", [True, False])
def test_create_programmatic_mapping_with_nested_partial_default(streaming):
    
    df = pl.DataFrame(
        {
            "MyValue": ["A", "B"],
            "MyOtherValue": ["321", None],
            "YetAnotherValue": [None, 321],
        }
    )
    mapping = Model()
    ex = Prefix("http://example.net/ns#")
    my_value = Variable("MyValue")
    my_other_value = Variable("MyOtherValue")
    yet_another_value = Variable("YetAnotherValue")
    my_object = ex.suf("MyObject")
    template = Template(
        ex.suf("ExampleTemplate"),
        [
            Parameter(my_value),
            Parameter(my_other_value, default_value=Literal("123")),
            Parameter(
                yet_another_value, default_value=Literal("123", data_type=xsd.long)
            ),
        ],
        [
            Triple(my_object, ex.suf("hasValue"), my_value),
            Triple(my_object, ex.suf("hasOtherValue"), my_other_value),
            Triple(my_object, ex.suf("hasYetAnotherValue"), yet_another_value),
        ],
    )
    template2 = Template(
        ex.suf("ExampleTemplate2"),
        [
            my_value,
            Parameter(my_other_value, optional=True),
            Parameter(yet_another_value, optional=True),
        ],
        [template.instance([my_value, my_other_value, yet_another_value])],
    )
    mapping.add_template(template)
    mapping.map(template2, df)
    qres = mapping.query(
        """
        PREFIX ex:<http://example.net/ns#>
        
        SELECT ?S ?A ?B WHERE {
        ?S ex:hasOtherValue ?A .
        ?S ex:hasYetAnotherValue ?B .
        } ORDER BY ?S ?A ?B
        """,
        streaming=streaming,
    )
    expected_df = pl.DataFrame(
        {
            "S": [f"<{my_object.iri}>"] * 4,
            "A": ["123", "123", "321", "321"],
            "B": [123, 321, 123, 321],
        }
    )
    assert_frame_equal(qres, expected_df)


@pytest.mark.parametrize("streaming", [True, False])
def test_create_programmatic_mapping_with_nested_none_optional(streaming):
    
    df = pl.DataFrame({"MyValue": ["A", "B"]})
    mapping = Model()
    ex = Prefix("http://example.net/ns#")
    my_value = Variable("MyValue")
    my_other_value = Variable("MyOtherValue")
    my_object = ex.suf("MyObject")
    template = Template(
        ex.suf("ExampleTemplate"),
        [
            my_value,
            Parameter(
                my_other_value, optional=True, rdf_type=RDFType.Literal(xsd.string)
            ),
        ],
        [
            Triple(my_object, ex.suf("hasValue"), my_value),
            Triple(my_object, ex.suf("hasOtherValue"), my_other_value),
        ],
    )
    template2 = Template(
        ex.suf("ExampleTemplate2"), [my_value], [template.instance([my_value, None])]
    )
    mapping.add_template(template)
    mapping.map(template2, df)
    qres = mapping.query(
        """
        PREFIX ex:<http://example.net/ns#>
        
        SELECT ?S ?A ?B WHERE {
        ?S ex:hasValue ?A .
        OPTIONAL {
            ?S ex:hasOtherValue ?B .
        }
        } ORDER BY ?S ?A ?B
        """,
        streaming=streaming,
    )
    expected_df = pl.DataFrame(
        {"S": [f"<{my_object.iri}>"] * 2, "A": ["A", "B"], "B": [None, None]}
    )
    expected_df = expected_df.with_columns(pl.col("B").cast(pl.String))
    assert_frame_equal(qres, expected_df)


@pytest.mark.parametrize("streaming", [True, False])
def test_create_programmatic_mapping_with_partial_default(streaming):
    
    df = pl.DataFrame(
        {
            "MyValue": ["A", "B"],
            "MyOtherValue": ["321", None],
            "YetAnotherValue": [None, 321],
        }
    )
    mapping = Model()
    ex = Prefix("http://example.net/ns#")
    my_value = Variable("MyValue")
    my_other_value = Variable("MyOtherValue")
    yet_another_value = Variable("YetAnotherValue")
    my_object = ex.suf("MyObject")
    template = Template(
        ex.suf("ExampleTemplate"),
        [
            Parameter(my_value),
            Parameter(my_other_value, default_value=Literal("123")),
            Parameter(
                yet_another_value, default_value=Literal("123", data_type=xsd.long)
            ),
        ],
        [
            Triple(my_object, ex.suf("hasValue"), my_value),
            Triple(my_object, ex.suf("hasOtherValue"), my_other_value),
            Triple(my_object, ex.suf("hasYetAnotherValue"), yet_another_value),
        ],
    )
    mapping.map(template, df)
    qres = mapping.query(
        """
        PREFIX ex:<http://example.net/ns#>
        
        SELECT ?S ?A ?B WHERE {
        ?S ex:hasOtherValue ?A .
        ?S ex:hasYetAnotherValue ?B .
        } ORDER BY ?S ?A ?B
        """,
        streaming=streaming,
    )
    expected_df = pl.DataFrame(
        {
            "S": [f"<{my_object.iri}>"] * 4,
            "A": ["123", "123", "321", "321"],
            "B": [123, 321, 123, 321],
        }
    )
    assert_frame_equal(qres, expected_df)


@pytest.mark.parametrize("streaming", [True, False])
def test_create_mapping_from_empty_signature(streaming):
    doc = """
    @prefix ex:<http://example.net/ns#>.
    ex:ExampleTemplate [] :: {
    ottr:Triple(ex:myObject, ex:hasObj, ex:myOtherObject)
    } .
    """

    mapping = Model()
    mapping.add_template(doc)
    mapping.map("http://example.net/ns#ExampleTemplate")
    qres = mapping.query(
        """
        PREFIX ex:<http://example.net/ns#>

        SELECT ?obj1 ?obj2 WHERE {
        ?obj1 ex:hasObj ?obj2
        } 
        """,
        streaming=streaming,
    )
    expected_df = pl.DataFrame(
        {
            "obj1": "<http://example.net/ns#myObject>",
            "obj2": "<http://example.net/ns#myOtherObject>",
        }
    )

    assert_frame_equal(qres, expected_df)


@pytest.mark.parametrize("streaming", [True, False])
def test_uri_subject_query(streaming):
    doc = """
    @prefix ex:<http://example.net/ns#>.
    ex:ExampleTemplate [] :: {
    ottr:Triple(ex:myObject, ex:hasObj, ex:myOtherObject)
    } .
    """

    model = Model()
    model.add_template(doc)
    model.map("http://example.net/ns#ExampleTemplate")
    qres = model.query(
        """
        PREFIX ex:<http://example.net/ns#>

        SELECT ?obj2 WHERE {
        ex:myObject ex:hasObj ?obj2
        } 
        """,
        streaming=streaming,
    )
    expected_df = pl.DataFrame({"obj2": "<http://example.net/ns#myOtherObject>"})

    assert_frame_equal(qres, expected_df)


def test_programmatic_model():
    example_template = Template(
        iri=IRI("http://example.net/ns#ExampleTemplate"),
        parameters=[],
        instances=[
            Triple(
                IRI("http://example.net/ns#myObject"),
                IRI("http://example.net/ns#hasObj"),
                IRI("http://example.net/ns#myOtherObject"),
            )
        ],
    )

    model = Model()
    model.add_template(example_template)
    model.map("http://example.net/ns#ExampleTemplate")
    qres = model.query(
        """
        PREFIX ex:<http://example.net/ns#>

        SELECT ?obj2 WHERE {
        ex:myObject ex:hasObj ?obj2
        } 
        """
    )
    expected_df = pl.DataFrame({"obj2": "<http://example.net/ns#myOtherObject>"})

    assert_frame_equal(qres, expected_df)


def test_programmatic_model_to_string():
    ex = Prefix("http://example.net/ns#", prefix_name="ex")
    example_template = Template(
        iri=ex.suf("ExampleTemplate"),
        parameters=[],
        instances=[
            Triple(
                ex.suf("myObject"),
                ex.suf("hasObj"),
                ex.suf("myOtherObject"),
            )
        ],
    )
    assert (
        str(example_template)
        == """<http://example.net/ns#ExampleTemplate> [ ] :: {
  <http://ns.ottr.xyz/0.4/Triple>(<http://example.net/ns#myObject>, <http://example.net/ns#hasObj>, <http://example.net/ns#myOtherObject>)
} . 
"""
    )


def test_programmatic_model_with_prefix():
    ex = Prefix("http://example.net/ns#", "ex")
    example_template = Template(
        iri=ex.suf("ExampleTemplate"),
        parameters=[],
        instances=[
            Triple(
                ex.suf("myObject"),
                ex.suf("hasObj"),
                ex.suf("myOtherObject"),
            )
        ],
    )

    model = Model()
    model.map(example_template)
    qres = model.query(
        """
        PREFIX ex:<http://example.net/ns#>

        SELECT ?obj2 WHERE {
        ex:myObject ex:hasObj ?obj2
        } 
        """
    )
    expected_df = pl.DataFrame({"obj2": "<http://example.net/ns#myOtherObject>"})

    assert_frame_equal(qres, expected_df)


def test_create_and_write_no_bug_string_lit_df():
    doc = """
    @prefix ex:<http://example.net/ns#>.
    ex:ExampleTemplate [?MyString] :: {
    ottr:Triple(ex:myObject, ex:hasValue, ?MyString)
    } .
    """

    df = pl.DataFrame({"MyString": ['"!!\n#"', "ABC#123\n\t\r"]})
    model = Model()
    model.add_template(doc)
    model.map("http://example.net/ns#ExampleTemplate", df)
    ntfile = "create_and_write_no_bug_string_lit.nt"
    model.write(ntfile, format="ntriples")
    with open(ntfile) as f:
        lines = f.readlines()

    lines.sort()
    assert lines == [
        '<http://example.net/ns#myObject> <http://example.net/ns#hasValue> "ABC#123\\n\\t\\r" .\n',
        '<http://example.net/ns#myObject> <http://example.net/ns#hasValue> "\\"!!\\n#\\"" .\n',
    ]


def test_create_and_write_no_bug_bool_lit_df():
    doc = """
    @prefix ex:<http://example.net/ns#>.
    ex:ExampleTemplate [?MyBool] :: {
    ottr:Triple(ex:myObject, ex:hasValue, ?MyBool)
    } .
    """

    df = pl.DataFrame({"MyBool": [True, False]})
    model = Model()
    model.add_template(doc)
    model.map("http://example.net/ns#ExampleTemplate", df)
    ntfile = "create_and_write_no_bug_bool_lit.nt"
    model.write(ntfile, format="ntriples")
    with open(ntfile) as f:
        lines = f.readlines()

    lines.sort()
    assert lines == [
        '<http://example.net/ns#myObject> <http://example.net/ns#hasValue> "false"^^<http://www.w3.org/2001/XMLSchema#boolean> .\n',
        '<http://example.net/ns#myObject> <http://example.net/ns#hasValue> "true"^^<http://www.w3.org/2001/XMLSchema#boolean> .\n',
    ]


def test_create_and_write_no_bug_lang_string_lit_df():
    model = Model()
    model.reads(
        '<http://example.net/ns#myObject> <http://example.net/ns#hasValue> "HELLO!!"@en .',
        format="ntriples",
    )
    ntfile = "create_and_write_no_bug_string_lit.nt"
    model.write(ntfile, format="ntriples")
    with open(ntfile) as f:
        lines = f.readlines()

    lines.sort()
    assert lines == [
        '<http://example.net/ns#myObject> <http://example.net/ns#hasValue> "HELLO!!"@en .\n'
    ]


def test_nested_template_empty_list():
    templates = """
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix owl: <http://www.w3.org/2002/07/owl#> .
<http://example.net/ns#ExampleNestedTemplate> [ ?MyValue ] :: {
  cross | <http://ns.ottr.xyz/0.4/Triple>(<http://example.net/ns#MyObject>,<http://example.net/ns#hasValue>,++?MyValue)
} . 
<http://example.net/ns#ExampleTemplate> [] :: {
   <http://example.net/ns#ExampleNestedTemplate>((1,2))
} . 
    """
    model = Model()
    model.add_template(templates)
    model.map("http://example.net/ns#ExampleTemplate")
    r = model.query(
        """
    SELECT ?a ?b ?c WHERE {
        ?a ?b ?c
    }
    """,
        include_datatypes=True,
    )
    assert r.rdf_types["c"] == RDFType.Literal(xsd.integer)


def test_bool_func():
    m = Model()
    df = m.query(
        """
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a WHERE {
    VALUES ?a {"true" "false"}
    FILTER(xsd:boolean(?a))
    }
    """
    )
    assert_frame_equal(df, pl.DataFrame({"a": ["true"]}))


def test_map_triples():
    df = pl.DataFrame(
        {
            "subject": [
                "http://example.net/ns#A",
                "http://example.net/ns#B",
                "http://example.net/ns#C",
            ],
            "object": [
                "http://example.net/ns#D",
                "http://example.net/ns#E",
                "http://example.net/ns#F",
            ],
        }
    )
    types = {"subject": RDFType.IRI, "object": RDFType.IRI}
    predicate = "http://example.net/ns#hasRel"
    m = Model()
    m.map_triples(df, types=types, predicate=predicate)
    df = m.query(
        """
    SELECT ?a ?b ?c WHERE {?a ?b ?c} ORDER BY ?a ?b ?c
    """
    )
    expected = pl.from_repr(
        """
┌───────────────────────────┬────────────────────────────────┬───────────────────────────┐
│ a                         ┆ b                              ┆ c                         │
│ ---                       ┆ ---                            ┆ ---                       │
│ str                       ┆ str                            ┆ str                       │
╞═══════════════════════════╪════════════════════════════════╪═══════════════════════════╡
│ <http://example.net/ns#A> ┆ <http://example.net/ns#hasRel> ┆ <http://example.net/ns#D> │
│ <http://example.net/ns#B> ┆ <http://example.net/ns#hasRel> ┆ <http://example.net/ns#E> │
│ <http://example.net/ns#C> ┆ <http://example.net/ns#hasRel> ┆ <http://example.net/ns#F> │
└───────────────────────────┴────────────────────────────────┴───────────────────────────┘
    """
    )
    assert_frame_equal(df, expected)


def test_map_default_triples():
    df = pl.DataFrame(
        {
            "subject": [
                "http://example.net/ns#A",
                "http://example.net/ns#B",
                "http://example.net/ns#C",
            ],
            "object": [
                "http://example.net/ns#D",
                "http://example.net/ns#E",
                "http://example.net/ns#F",
            ],
        }
    )
    m = Model()
    tpl = m.map_default(df, primary_key_column="subject")
    # print(tpl)
    df = m.query(
        """
    SELECT ?a ?b ?c WHERE {?a ?b ?c} ORDER BY ?a ?b ?c
    """
    )
    expected = pl.from_repr(
        """
┌───────────────────────────┬────────────────────────────────┬───────────────────────────┐
│ a                         ┆ b                              ┆ c                         │
│ ---                       ┆ ---                            ┆ ---                       │
│ str                       ┆ str                            ┆ str                       │
╞═══════════════════════════╪════════════════════════════════╪═══════════════════════════╡
│ <http://example.net/ns#A> ┆ <urn:maplib_default:object>    ┆ <http://example.net/ns#D> │
│ <http://example.net/ns#B> ┆ <urn:maplib_default:object>    ┆ <http://example.net/ns#E> │
│ <http://example.net/ns#C> ┆ <urn:maplib_default:object>    ┆ <http://example.net/ns#F> │
└───────────────────────────┴────────────────────────────────┴───────────────────────────┘
    """
    )
    assert_frame_equal(df, expected)


def test_map_default_triples_non_iri_object():
    df = pl.DataFrame(
        {
            "subject": [
                "http://example.net/ns#A",
                "http://example.net/ns#B",
                "http://example.net/ns#C",
            ],
            "object": [
                "1",
                "3",
                "5",
            ],
        }
    )
    m = Model()
    tpl = m.map_default(df, primary_key_column="subject")
    # print(tpl)
    df = m.query(
        """
    SELECT ?a ?b ?c WHERE {?a ?b ?c} ORDER BY ?a ?b ?c
    """
    )
    expected = pl.from_repr(
        """
┌───────────────────────────┬────────────────────────────────┬───────────────────────────┐
│ a                         ┆ b                              ┆ c                         │
│ ---                       ┆ ---                            ┆ ---                       │
│ str                       ┆ str                            ┆ str                       │
╞═══════════════════════════╪════════════════════════════════╪═══════════════════════════╡
│ <http://example.net/ns#A> ┆ <urn:maplib_default:object>    ┆ 1                         │
│ <http://example.net/ns#B> ┆ <urn:maplib_default:object>    ┆ 3                         │
│ <http://example.net/ns#C> ┆ <urn:maplib_default:object>    ┆ 5                         │
└───────────────────────────┴────────────────────────────────┴───────────────────────────┘
    """
    )
    assert_frame_equal(df, expected)


def test_map_generated_default_triples_non_iri_object():
    df = pl.DataFrame(
        {
            "subject": [
                "http://example.net/ns#A",
                "http://example.net/ns#B",
                "http://example.net/ns#C",
            ],
            "object": [
                "1",
                "3",
                "5",
            ],
        }
    )
    m = Model()
    tpl = """<urn:maplib_default:default_template_0> [ <http://ns.ottr.xyz/0.4/IRI> ?subject,  <http://www.w3.org/2001/XMLSchema#string> ?object ] :: {
                ottr:Triple(?subject,<urn:maplib_default:object>,?object)
            } . """
    m.map(tpl, df)
    df = m.query(
        """
    SELECT ?a ?b ?c WHERE {?a ?b ?c} ORDER BY ?a ?b ?c
    """
    )
    expected = pl.from_repr(
        """
┌───────────────────────────┬────────────────────────────────┬───────────────────────────┐
│ a                         ┆ b                              ┆ c                         │
│ ---                       ┆ ---                            ┆ ---                       │
│ str                       ┆ str                            ┆ str                       │
╞═══════════════════════════╪════════════════════════════════╪═══════════════════════════╡
│ <http://example.net/ns#A> ┆ <urn:maplib_default:object>    ┆ 1                         │
│ <http://example.net/ns#B> ┆ <urn:maplib_default:object>    ┆ 3                         │
│ <http://example.net/ns#C> ┆ <urn:maplib_default:object>    ┆ 5                         │
└───────────────────────────┴────────────────────────────────┴───────────────────────────┘
    """
    )
    assert_frame_equal(df, expected)



def test_list_expansion_correct():
    df = pl.DataFrame(
        {
            "MySubject": [
                "http://example.net/ns#subject1",
                "http://example.net/ns#subject2",
            ],
            "MyValue": [[1, 2], [3, 4]],
        }
    )
    templates = """
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix owl: <http://www.w3.org/2002/07/owl#> .
<http://example.net/ns#ExampleTemplate> [ ottr:IRI ?MySubject, List<rdfs:Resource> ?MyValue ] :: {
  ottr:Triple(?MySubject,<http://example.net/ns#hasValueList>,?MyValue)
} . 
    """
    model = Model()
    model.add_template(templates)
    model.map("http://example.net/ns#ExampleTemplate", df)
    df = model.query(
        """
    PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> 
    SELECT ?a ?e1 ?e2 ?e3 WHERE {
        ?a <http://example.net/ns#hasValueList> ?b .
        ?b rdf:first ?e1 .
        ?b rdf:rest ?c .
        ?c rdf:first ?e2 .
        ?c rdf:rest ?e3 .
    } ORDER BY ?a
    """
    )
    expected = pl.from_repr(
        """
┌──────────────────────────────────┬─────┬─────┬──────────────────────────────────────────────────┐
│ a                                ┆ e1  ┆ e2  ┆ e3                                               │
│ ---                              ┆ --- ┆ --- ┆ ---                                              │
│ str                              ┆ i64 ┆ i64 ┆ str                                              │
╞══════════════════════════════════╪═════╪═════╪══════════════════════════════════════════════════╡
│ <http://example.net/ns#subject1> ┆ 1   ┆ 2   ┆ <http://www.w3.org/1999/02/22-rdf-syntax-ns#nil> │
│ <http://example.net/ns#subject2> ┆ 3   ┆ 4   ┆ <http://www.w3.org/1999/02/22-rdf-syntax-ns#nil> │
└──────────────────────────────────┴─────┴─────┴──────────────────────────────────────────────────┘
    """
    )
    assert_frame_equal(df, expected)



# r'''
# # Overview
#
# '''

__all__ = [
    "Model",
    "a",
    "Triple",
    "SolutionMappings",
    "IndexingOptions",
    "ValidationReport",
    "Instance",
    "Template",
    "Argument",
    "Parameter",
    "Variable",
    "RDFType",
    "xsd",
    "rdf",
    "rdfs",
    "owl",
    "IRI",
    "Literal",
    "Prefix",
    "BlankNode",
    "explore",
    "add_triples",
    "generate_templates",
    "MaplibException",
]

import pathlib
from importlib.metadata import version
from .maplib import *
from .adding_triples import add_triples
from .template_generator import generate_templates

"""
http://www.w3.org/1999/02/22-rdf-syntax-ns#type
"""
a = rdf.type

if (pathlib.Path(__file__).parent.resolve() / "graph_explorer").exists():
    from .graph_explorer import explore
else:

    def explore(
        m: "Model",
        host: str = "localhost",
        port: int = 8000,
        bind: str = "localhost",
        popup=True,
        fts=True,
        fts_path:str="fts",
    ):
        """Starts a graph explorer session.
        To run from Jupyter Notebook use:
        >>> from maplib import explore
        >>>
        >>> server = explore(m)
        You can later stop the server with
        >>> server.stop()

        :param m: The Model to explore
        :param host: The hostname that we will point the browser to.
        :param port: The port where the graph explorer webserver listens on.
        :param bind: Bind to the following host / ip.
        :param popup: Pop up the browser window.
        :param fts: Enable full text search indexing
        :param fts_path: Path to the fts index
        """
        print("Contact Data Treehouse to try!")

__version__ = version("maplib") 



import pytest
from maplib import Model, Triple, Variable, Template, IRI, Parameter, RDFType
import polars as pl


def test_many_validation_errors_subject():
    n = 10_000
    df = pl.DataFrame(
        {
            "subject": [f"MyBadIRI//{i}" for i in range(n)],
            "predicate": ["https://example.net/hasObject"] * n,
            "object": [i for i in range(n)],
        }
    )
    m = Model()
    subj = Variable("subject")
    pred = Variable("predicate")
    obj = Variable("object")
    t = Template(
        IRI("https://github.com/DataTreehouse/maplib/my_template"),
        [subj, pred, obj],
        [Triple(subj, pred, obj)],
    )
    with pytest.raises(Exception) as e:
        m.map(t, df, validate_iris=True)
    assert "invalid IRI" in str(e)


def test_many_validation_errors_predicate():
    n = 10_000
    df = pl.DataFrame(
        {
            "subject": [f"urn:dth:goodiri_{i}" for i in range(n)],
            "predicate": ["!https://example.net/hasObject"] * n,
            "object": [i for i in range(n)],
        }
    )
    m = Model()
    subj = Variable("subject")
    pred = Variable("predicate")
    obj = Variable("object")
    t = Template(
        IRI("https://github.com/DataTreehouse/maplib/my_template"),
        [subj, pred, obj],
        [Triple(subj, pred, obj)],
    )
    with pytest.raises(Exception) as e:
        m.map(t, df, validate_iris=True)
    assert "invalid IRI" in str(e)


def test_many_validation_errors_nested_object():
    n = 1_000
    df = pl.DataFrame(
        {
            "subject": [f"urn:dth:goodiri_{i}" for i in range(n)],
            "predicate": ["https://example.net/hasObject"] * n,
            "object": [
                [f"aTerribleIRI!!{i}", f"anotherTerribleIRI!!{i}"] for i in range(n)
            ],
        }
    )
    m = Model()
    subj = Variable("subject")
    pred = Variable("predicate")
    obj = Variable("object")
    t = Template(
        IRI("https://github.com/DataTreehouse/maplib/my_template"),
        [subj, pred, Parameter(obj, rdf_type=RDFType.Nested(RDFType.IRI))],
        [Triple(subj, pred, obj)],
    )
    with pytest.raises(Exception) as e:
        m.map(t, df, validate_iris=True)
    assert "invalid IRI" in str(e)



from pathlib import Path
from typing import Union, List, Dict, Optional, Callable, Tuple, Literal as LiteralType
from polars import DataFrame
from datetime import datetime, date
from maplib.maplib import rdf

a = rdf.type

class xsd:
    """
    The xsd namespace: http://www.w3.org/2001/XMLSchema#
    """
    anyURI:IRI
    base64Binary:IRI
    boolean:IRI
    byte:IRI
    date:IRI
    dateTime:IRI
    dateTimeStamp:IRI
    dayTimeDuration:IRI
    decimal:IRI
    double:IRI
    duration:IRI
    float:IRI
    gDay:IRI
    gMonth:IRI
    gMonthDay:IRI
    gYear:IRI
    gYearMonth:IRI
    hexBinary:IRI
    int_:IRI
    integer:IRI
    language:IRI
    long:IRI
    negativeInteger:IRI
    nonNegativeInteger:IRI
    nonPositiveInteger:IRI
    normalizedString:IRI
    short:IRI
    string:IRI
    token:IRI
    unsignedByte:IRI
    unsignedInt:IRI
    unsignedLong:IRI
    unsignedShort:IRI
    yearMonthDuration:IRI


class rdf:
    """
    The rdf namespace: http://www.w3.org/1999/02/22-rdf-syntax-ns#
    """
    Alt:IRI
    Bag:IRI
    HTML:IRI
    List:IRI
    Property:IRI
    Seq:IRI
    Statement:IRI
    XMLLiteral:IRI
    first:IRI
    langString:IRI
    nil:IRI
    object:IRI
    predicate:IRI
    rest:IRI
    subject:IRI
    type:IRI
    value:IRI


class rdfs:
    """
    The rdfs namespace: http://www.w3.org/1999/02/22-rdf-syntax-ns#
    """
    Class:IRI
    Container:IRI
    ContainerMembershipProperty:IRI
    Datatype:IRI
    Literal:IRI
    Resource:IRI
    comment:IRI
    domain:IRI
    isDefinedBy:IRI
    label:IRI
    member:IRI
    range:IRI
    seeAlso:IRI
    subClassOf:IRI
    subPropertyOf:IRI


class owl:
    """
    The owl namespace: http://www.w3.org/2002/07/owl#
    """
    AllDifferent:IRI
    AllDisjointClasses:IRI
    AllDisjointProperties:IRI
    Annotation:IRI
    AnnotationProperty:IRI
    Axiom:IRI
    Class:IRI
    DataRange:IRI
    DatatypeProperty:IRI
    DeprecatedClass:IRI
    DeprecatedProperty:IRI
    FunctionalProperty:IRI
    InverseFunctionalProperty:IRI
    IrreflexiveProperty:IRI
    NamedIndividual:IRI
    NegativePropertyAssertion:IRI
    Nothing:IRI
    ObjectProperty:IRI
    Ontology:IRI
    OntologyProperty:IRI
    ReflexiveProperty:IRI
    Restriction:IRI
    SymmetricProperty:IRI
    Thing:IRI
    TransitiveProperty:IRI
    allValuesFrom:IRI
    annotatedProperty:IRI
    annotatedSource:IRI
    annotatedTarget:IRI
    assertionProperty:IRI
    backwardCompatibleWith:IRI
    bottomDataProperty:IRI
    bottomObjectProperty:IRI
    cardinality:IRI
    complementOf:IRI
    datatypeComplementOf:IRI
    deprecated:IRI
    differentFrom:IRI
    disjointUnionOf:IRI
    disjointWith:IRI
    distinctMembers:IRI
    equivalentClass:IRI
    equivalentProperty:IRI
    hasKey:IRI
    hasSelf:IRI
    hasValue:IRI
    imports:IRI
    incompatibleWith:IRI
    intersectionOf:IRI
    inverseOf:IRI
    maxCardinality:IRI
    maxQualifiedCardinality:IRI
    members:IRI
    minCardinality:IRI
    minQualifiedCardinality:IRI
    onClass:IRI
    onDataRange:IRI
    onDatatype:IRI
    onProperties:IRI
    onProperty:IRI
    oneOf:IRI
    priorVersion:IRI
    propertyChainAxiom:IRI
    propertyDisjointWith:IRI
    qualifiedCardinality:IRI
    rational:IRI
    real:IRI
    sameAs:IRI
    someValuesFrom:IRI
    sourceIndividual:IRI
    targetIndividual:IRI
    targetValue:IRI
    topDataProperty:IRI
    topObjectProperty:IRI
    unionOf:IRI
    versionIRI:IRI
    versionInfo:IRI
    withRestrictions:IRI


class RDFType:
    """
    The type of a column containing a RDF variable.
    For instance, IRIs are RDFType.IRI and a string literal is RDFType.Literal("http://www.w3.org/2001/XMLSchema#string")
    """

    IRI: "RDFType"
    BlankNode: "RDFType"
    Literal: Callable[[Union[str, "IRI"]], "RDFType"]
    Multi: Callable[[List["RDFType"]], "RDFType"]
    Nested: Callable[["RDFType"], "RDFType"]
    Unknown: "RDFType"

class SolutionMappings:
    """
    Detailed information about the solution mappings, the types of the variables and debugging for queries.
    """

    mappings: DataFrame
    rdf_types: Dict[str, RDFType]
    debug: Optional[str]

class Variable:
    """
    A variable in a template.
    """

    name: str

    def __init__(self, name: str):
        """
        Create a new variable.
        :param name: The name of the variable.
        """
        ...

class IRI:
    iri: str
    """
    An IRI.
    """

    def __init__(self, iri: str):
        """
        Create a new IRI
        :param iri: IRI (without < and >).
        """

class BlankNode:
    """
    A Blank Node.
    """

    name: str

    def __init__(self, name: str):
        """
        Create a new Blank Node
        :param name: Name of blank node (without _:).
        """

class Prefix:
    """
    A prefix that can be used to ergonomically build iris.
    """

    def __init__(self, iri,  prefix_name=None):
        """
        Create a new prefix.
        :param iri: The prefix IRI.
        :param prefix_name: The name of the prefix
        """

    def suf(self, suffix: str) -> IRI:
        """
        Create an IRI by appending the suffix.
        :param suffix: The suffix to append.
        :return:
        """

class Literal:
    """
    An RDF literal.
    """

    value: str
    datatype: Optional[IRI]
    language: Optional[str]

    def __init__(self, value: str, data_type: IRI = None, language: str = None):
        """
        Create a new RDF Literal
        :param value: The lexical representation of the value.
        :param data_type: The data type of the value (an IRI).
        :param language: The language tag of the value.
        """

    def to_native(self) -> Union[int, float, bool, str, datetime, date]:
        """

        :return:
        """

class Parameter:
    variable: Variable
    optional: bool
    allow_blank: bool
    rdf_type: Optional[RDFType]
    default_value: Optional[Union[Literal, IRI, BlankNode]]
    """
    Parameters for template signatures.
    """

    def __init__(
        self,
        variable: Variable,
        optional: Optional[bool] = False,
        allow_blank: Optional[bool] = True,
        rdf_type: Optional[RDFType] = None,
        default_value: Optional[Union[Literal, IRI, BlankNode]] = None,
    ):
        """
        Create a new parameter for a Template.
        :param variable: The variable.
        :param optional: Can the variable be unbound?
        :param allow_blank: Can the variable be bound to a blank node?
        :param rdf_type: The type of the variable. Can be nested.
        :param default_value: Default value when no value provided.
        """

class Argument:
    def __init__(
        self, term: Union[Variable, IRI, Literal], list_expand: Optional[bool] = False
    ):
        """
        An argument for a template instance.
        :param term: The term.
        :param list_expand: Should the argument be expanded? Used with the list_expander argument of instance.
        """

class Instance:
    def __init__(
        self,
        iri: IRI,
        arguments: List[Union[Argument, Variable, IRI, Literal, BlankNode, None]],
        list_expander: Optional[LiteralType["cross", "zipMin", "zipMax"]] = None,
    ):
        """
        A template instance.
        :param iri: The IRI of the template to be instantiated.
        :param arguments: The arguments for template instantiation.
        :param list_expander: (How) should we do list expansion?
        """

class Template:
    iri: str
    parameters: List[Parameter]
    instances: List[Instance]
    """
    An OTTR Template.
    Note that accessing parameters- or instances-fields returns copies. 
    To change these fields, you must assign new lists of parameters or instances.  
    """

    def __init__(
        self,
        iri: IRI,
        parameters: List[Union[Parameter, Variable]],
        instances: List[Instance],
    ):
        """
        Create a new OTTR Template
        :param iri: The IRI of the template
        :param parameters:
        :param instances:
        """

    def instance(
        self,
        arguments: List[Union[Argument, Variable, IRI, Literal, None]],
        list_expander: LiteralType["cross", "zipMin", "zipMax"] = None,
    ) -> Instance:
        """

        :param arguments: The arguments to the template.
        :param list_expander: (How) should we list-expand?
        :return:
        """

def Triple(
    subject: Union["Argument", IRI, Variable, BlankNode],
    predicate: Union["Argument", IRI, Variable, BlankNode],
    object: Union["Argument", IRI, Variable, Literal, BlankNode],
    list_expander: Optional[LiteralType["cross", "zipMin", "zipMax"]] = None,
):
    """
    An OTTR Triple Pattern used for creating templates.
    This is the basis pattern which all template instances are rewritten into.
    Equivalent to:

    >>> ottr = Prefix("http://ns.ottr.xyz/0.4/")
    ... Instance(ottr.suf("Triple"), subject, predicate, object, list_expander)

    :param subject:
    :param predicate:
    :param object:
    :param list_expander:
    :return:
    """

class XSD:
    """
    The xsd namespace, for convenience.
    """

    boolean: IRI
    byte: IRI
    date: IRI
    dateTime: IRI
    dateTimeStamp: IRI
    decimal: IRI
    double: IRI
    duration: IRI
    float: IRI
    int_: IRI
    integer: IRI
    language: IRI
    long: IRI
    short: IRI
    string: IRI

    def __init__(self):
        """
        Create the xsd namespace helper.
        """

# END COMMON WITH CHRONTEXT

class IndexingOptions:
    """
    Options for indexing
    """

    def __init__(
        self,
        object_sort_all: bool = None,
        object_sort_some: List["IRI"] = None,
        fts_path: str = None,
    ):
        """
        Defaults to indexing on subjects and objects for select types (e.g. rdf:type and rdfs:label)

        :param object_sort_all: Enable object-indexing for all suitable predicates (doubles memory requirement).
        :param object_sort_some: Enable object-indexing for a selected list of predicates.
        :param fts_path: Enable full text search, stored at the path
        """

ParametersType = Dict[str, Tuple[DataFrame, Dict[str, RDFType]]]

class ValidationReport:
    """
    SHACL Validation report.
    Only constructed by maplib.
    """

    conforms: bool
    "Whether or not the validation report conforms to the shapes"

    shape_targets: DataFrame
    "A DataFrame containing the counts of the targets of each shape and constraint"

    performance: DataFrame
    "Performance statistics for the validation process"

    def results(
        self,
        native_dataframe: bool = False,
        include_datatypes: bool = False,
        streaming: bool = False,
    ) -> Optional[Union[DataFrame, "SolutionMappings"]]:
        """
        Return the results of the validation report, if they exist.

        :param native_dataframe: Return columns with maplib-native formatting. Useful for round-trips.
        :param include_datatypes: Return datatypes of the results DataFrame (returns SolutionMappings instead of DataFrame).
        :param streaming: Use the Polars streaming functionality.
        :return: The SHACL validation report, as a DataFrame
        """

    def details(
        self,
        native_dataframe: bool = False,
        include_datatypes: bool = False,
        streaming: bool = False,
    ) -> Optional[DataFrame]:
        """
        Returns the details of the validation report.
        Only available if validation was called with include_details=True.

        :param native_dataframe: Return columns with maplib-native formatting. Useful for round-trips.
        :param include_datatypes: Return datatypes of the results DataFrame (returns SolutionMappings instead of DataFrame).
        :param streaming: Use the Polars streaming functionality.
        :return: Details of the SHACL validation report, as a DataFrame
        """

    def graph(self) -> "Mapping":
        """
        Creates a new model object where the base graph is the validation report with results.
        Includes the details of the validation report in the new graph if they exist.

        :return:
        """

class Model:
    """
    A model session allowing:

    * Iterative model using OTTR templates
    * Interactive SPARQL querying and enrichment
    * SHACL validation

    Usage:

    >>> from maplib import Model
    ... doc = '''
    ... @prefix ex:<http://example.net/ns#>.
    ... ex:ExampleTemplate [?MyValue] :: {
    ...    ottr:Triple(ex:myObject, ex:hasValue, ?MyValue)
    ... } .'''
    ... m = Model()
    ... m.add_template(doc)

    :param documents: a stOTTR document or a list of these
    :param indexing_options: options for indexing
    """

    def __init__(
        self,
        indexing_options: "IndexingOptions" = None,
    ) -> "Model": ...

    def add_template(self, template: Union["Template", str]):
        """
        Add a template to the model. Overwrites any existing template with the same IRI.
        :param template: The template to add, as a stOTTR string or as a programmatically constructed Template.
        :return:
        """

    def add_prefixes(self, prefixes: Dict[str, str]):
        """
        Add prefixes that will be used in parsing of SPARQL, Datalog and OTTR.

        Usage:
        >>> m.add_prefixes({"ex" : "http:://example.net/"})

        :param prefixes: Known prefixes
        :return:
        """

    def map(
        self,
        template: Union[str, "Template", IRI],
        df: DataFrame = None,
        graph: str = None,
        types: Dict[str, RDFType] = None,
        validate_iris: bool = True,
    ) -> None:
        """
        Map a template using a DataFrame
        Usage:

        >>> m.map("ex:ExampleTemplate", df)

        If the template has no arguments, the df argument is not necessary.

        :param template: Template, IRI, IRI string or prefixed template name.
        :param df: DataFrame where the columns have the same names as the template arguments
        :param graph: The IRI of the graph to add triples to.
        :param types: The types of the columns.
        :param validate_iris: Validate any IRI-columns.
        """

    def map_triples(
        self,
        df: DataFrame = None,
        predicate: str = None,
        graph: str = None,
        types: Dict[str, RDFType] = None,
        validate_iris: bool = True,
    ) -> None:
        """
        Map a template using a DataFrame with columns subject, object and predicate
        The predicate column can also be supplied as a string if it is the same for all rows.
        Usage:

        >>> m.map_triples(df)

        If the template has no arguments, the df argument is not necessary.

        :param df: DataFrame where the columns are named subject and object. May also contain a verb-column.
        :param verb: The uri of the verb.
        :param graph: The IRI of the graph to add triples to.
        :param types: The types of the columns.
        :param validate_iris: Validate any IRI-columns.
        """

    def map_default(
        self,
        df: DataFrame,
        primary_key_column: str,
        dry_run: bool = False,
        graph: str = None,
        types: Dict[str, RDFType] = None,
        validate_iris: bool = True,
    ) -> str:
        """
        Create a default template and map it based on a dataframe.
        Usage:

        >>> template_string = m.map_default(df, "myKeyCol")
        ... print(template_string)

        :param df: DataFrame where the columns have the same names as the template arguments
        :param primary_key_column: This column will be the subject of all triples in the generated template.
        :param dry_run: Do not map the template, only return the string.
        :param graph: The IRI of the graph to add triples to.
        :param types: The types of the columns.
        :param validate_iris: Validate any IRI-columns.
        :return: The generated template
        """

    def query(
        self,
        query: str,
        parameters: ParametersType = None,
        include_datatypes: bool = False,
        native_dataframe: bool = False,
        graph: str = None,
        streaming: bool = False,
        return_json: bool = False,
        include_transient: bool = True,
        max_rows: int = None,
        debug: bool = False,
    ) -> Union[
        DataFrame, SolutionMappings, List[Union[DataFrame, SolutionMappings, str]], None
    ]:
        """
        Query the contained knowledge graph using SPARQL
        Currently, SELECT, CONSTRUCT and INSERT are supported.
        Usage:

        >>> df = model.query('''
        ... PREFIX ex:<http://example.net/ns#>
        ... SELECT ?obj1 ?obj2 WHERE {
        ...    ?obj1 ex:hasObj ?obj2
        ... }''')
        ... print(df)

        :param query: The SPARQL query string
        :param parameters: PVALUES Parameters, a DataFrame containing the value bindings in the custom PVALUES construction.
        :param native_dataframe: Return columns with maplib-native formatting. Useful for round-trips.
        :param include_datatypes: Datatypes are not returned by default, set to true to return a dict with the solution mappings and the datatypes.
        :param graph: The IRI of the graph to query.
        :param streaming: Use Polars streaming
        :param return_json: Return JSON string.
        :param include_transient: Include transient triples when querying.
        :param max_rows: Maximum estimated rows in result, helps avoid out-of-memory errors.
        :param debug: Why does my query have no results?
        :return: DataFrame (Select), list of DataFrames (Construct) containing results, None for Insert-queries, or SolutionMappings when debug or native_dataframe is set.

        """

    def update(
            self,
            update: str,
            parameters: ParametersType = None,
            streaming: bool = False,
            include_transient: bool = True,
            max_rows: int = None,
            debug: bool = False,
    ):
        """
        Insert the results of a Construct query in the graph.
        Useful for being able to use the same query for inspecting what will be inserted and actually inserting.
        Usage:

        >>> m = Model(doc)
        ... # Omitted
        ... update_pizzas = '''
        ... ...'''
        ... m.update(update_pizzas)

        :param update: The SPARQL Update string
        :param parameters: PVALUES Parameters, a DataFrame containing the value bindings in the custom PVALUES construction.
        :param streaming: Use Polars streaming
        :param include_transient: Include transient triples when querying (but see "transient" above).
        :param max_rows: Maximum estimated rows in result, helps avoid out-of-memory errors.
        :param debug: Why does my query have no results?
        :return: None
        """

    def insert(
        self,
        query: str,
        parameters: ParametersType = None,
        include_datatypes: bool = False,
        native_dataframe: bool = False,
        transient: bool = False,
        streaming: bool = False,
        source_graph: str = None,
        target_graph: str = None,
        include_transient: bool = True,
        max_rows: int = None,
        debug: bool = False,
    ):
        """
        Insert the results of a Construct query in the graph.
        Useful for being able to use the same query for inspecting what will be inserted and actually inserting.
        Usage:

        >>> m = Model(doc)
        ... # Omitted
        ... hpizzas = '''
        ... PREFIX pizza:<https://github.com/magbak/maplib/pizza#>
        ... PREFIX ing:<https://github.com/magbak/maplib/pizza/ingredients#>
        ... CONSTRUCT { ?p a pizza:HeterodoxPizza }
        ... WHERE {
        ... ?p a pizza:Pizza .
        ... ?p pizza:hasIngredient ing:Pineapple .
        ... }'''
        ... m.insert(hpizzas)

        :param query: The SPARQL Insert query string
        :param parameters: PVALUES Parameters, a DataFrame containing the value bindings in the custom PVALUES construction.
        :param native_dataframe: Return columns with maplib-native formatting. Useful for round-trips.
        :param include_datatypes: Datatypes are not returned by default, set to true to return a dict with the solution mappings and the datatypes.
        :param transient: Should the inserted triples be transient?
        :param source_graph: The IRI of the source graph to execute the construct query.
        :param target_graph: The IRI of the target graph to insert into.
        :param streaming: Use Polars streaming
        :param include_transient: Include transient triples when querying (but see "transient" above).
        :param max_rows: Maximum estimated rows in result, helps avoid out-of-memory errors.
        :param debug: Why does my query have no results?
        :return: None
        """

    def validate(
        self,
        shape_graph: str,
        data_graph: str = None,
        include_details: bool = False,
        include_conforms: bool = False,
        include_shape_graph: bool = True,
        streaming: bool = False,
        max_shape_constraint_results: int = None,
        only_shapes: List[str] = None,
        deactivate_shapes: List[str] = None,
        dry_run: bool = False,
        max_rows: int = None,
        serial: bool = False,
    ) -> ValidationReport:
        """
        Validate the contained knowledge graph using SHACL
        Assumes that the contained knowledge graph also contains SHACL Shapes.

        :param shape_graph: The IRI of the Shape Graph.
        :param data_graph: The IRI of the Data Graph (defaults to the default graph).
        :param include_details: Include details of SHACL evaluation alongside the report. Currently uses a lot of memory.
        :param include_conforms: Include those results that conformed. Also applies to details.
        :param include_shape_graph: Include the shape graph in the report, useful when creating the graph from the report.
        :param include_datatypes: Return the datatypes of the validation report (and details).
        :param streaming: Use Polars streaming
        :param max_shape_constraint_results: Maximum number of results per shape and constraint. Reduces the size of the result set.
        :param only_shapes: Validate only these shapes, None means all shapes are validated (must be IRI, cannot be used with deactivate_shapes).
        :param deactivate_shapes: Disable validation of these shapes (must be IRI, cannot be used with deactivate_shapes).
        :param dry_run: Only find targets of shapes, but do not validate them.
        :param max_rows: Maximum estimated rows in underlying SPARQL results, helps avoid out-of-memory errors.
        :param serial: Turns off most parallell validation of shapes.
        :return: Validation report containing a report (report.df) and whether the graph conforms (report.conforms)
        """

    def read(
        self,
        file_path: Union[str, Path],
        format: LiteralType["ntriples", "turtle", "rdf/xml", "xml", "rdfxml"] = None,
        base_iri: str = None,
        transient: bool = False,
        parallel: bool = None,
        checked: bool = True,
        graph: str = None,
        replace_graph: bool = False,
    ) -> None:
        """
        Reads triples from a file path.
        You can specify the format, or it will be derived using file extension, e.g. filename.ttl or filename.nt.
        Specify transient if you only want the triples to be available for further querying and validation,
        but not persisted using write-methods.

        Usage:

        >>> m.read("my_triples.ttl")

        :param file_path: The path of the file containing triples
        :param format: One of "ntriples", "turtle", "rdf/xml", otherwise it is inferred from the file extension.
        :param base_iri: Base iri
        :param transient: Should these triples be included when writing the graph to the file system?
        :param parallel: Parse triples in parallel, currently only NTRiples and Turtle. Assumes all prefixes are in the beginning of the document. Defaults to true only for NTriples.
        :param checked: Check IRIs etc.
        :param graph: The IRI of the graph to read the triples into, if None, it will be the default graph.
        :param replace_graph: Replace the graph with these triples? Will replace the default graph if no graph is specified.
        """

    def read_template(
            self,
            file_path: Union[str, Path],
    ) -> None:
        """
        Reads template(s) from a file path.

        Usage:

        >>> m.read("templates.ttl")

        :param file_path: The path of the file containing templates in stOTTR format
        """

    def reads(
        self,
        s: str,
        format: LiteralType["ntriples", "turtle", "rdf/xml", "xml", "rdfxml"],
        base_iri: str = None,
        transient: bool = False,
        parallel: bool = None,
        checked: bool = True,
        graph: str = None,
        replace_graph: bool = False,
    ) -> None:
        """
        Reads triples from a string.
        Specify transient if you only want the triples to be available for further querying and validation,
        but not persisted using write-methods.

        Usage:

        >>> m.reads(my_ntriples_string, format="ntriples")

        :param s: String containing serialized triples.
        :param format: One of "ntriples", "turtle", "rdf/xml".
        :param base_iri: Base iri
        :param transient: Should these triples be included when writing the graph to the file system?
        :param parallel: Parse triples in parallel, currently only NTRiples and Turtle. Assumes all prefixes are in the beginning of the document. Defaults to true for NTriples.
        :param checked: Check IRIs etc.
        :param graph: The IRI of the graph to read the triples into.
        :param replace_graph: Replace the graph with these triples? Will replace the default graph if no graph is specified.
        """

    def write_cim_xml(
        self,
        file_path: Union[str, Path],
        profile_graph: str,
        model_iri: str = None,
        version: str = None,
        description: str = None,
        created: str = None,
        scenario_time: str = None,
        modeling_authority_set: str = None,
        prefixes: Dict[str, str] = None,
        graph: str = None,
    ) -> None:
        """
        Write the legacy CIM XML format.

        >>> PROFILE_GRAPH = "urn:graph:profiles"
        >>> m = Model()
        >>> m.read(model_path, base_iri=publicID, format="rdf/xml")
        >>> m.read("61970-600-2_Equipment-AP-Voc-RDFS2020_v3-0-0.rdf", graph=PROFILE_GRAPH, format="rdf/xml")
        >>> m.read("61970-600-2_Operation-AP-Voc-RDFS2020_v3-0-0.rdf", graph=PROFILE_GRAPH, format="rdf/xml")
        >>> m.write_cim_xml(
        >>>     "model.xml",
        >>>     profile_graph=PROFILE_GRAPH,
        >>>     description = "MyModel",
        >>>     created = "2023-09-14T20:27:41",
        >>>     scenario_time = "2023-09-14T02:44:43",
        >>>     modeling_authority_set="www.westernpower.co.uk",
        >>>     version="22",
        >>> )

        :param file_path: The path of the file containing triples
        :param profile_graph: The IRI of the graph containing the ontology of the CIM profile to write.
        :param model_iri: model_iri a md:FullModel. Is generated if not provided.
        :param version: model_iri md:Model.version version .
        :param description: model_iri md:Model.description description .
        :param created: model_iri md:Model.created created .
        :param scenario_time: model_iri md:Model.scenarioTime scenario_time .
        :param modeling_authority_set: model_iri md:Model.modelingAuthoritySet modeling_authority_set .
        :param prefixes: Prefixes to be used in XML export.
        :param graph: The graph to write, defaults to the default graph.
        """

    def write(
        self,
        file_path: Union[str, Path],
        format=LiteralType["ntriples", "turtle", "rdf/xml"],
        graph: str = None,
    ) -> None:
        """
        Write the non-transient triples to the file path specified in the NTriples format.

        Usage:

        >>> m.write("my_triples.nt", format="ntriples")

        :param file_path: The path of the file containing triples
        :param format: One of "ntriples", "turtle", "rdf/xml".
        :param graph: The IRI of the graph to write.
        """

    def writes(
        self, format=LiteralType["ntriples", "turtle", "rdf/xml"], graph: str = None
    ) -> str:
        """
        DEPRECATED: use writes with format="ntriples"
        Write the non-transient triples to a string in memory.

        Usage:

        >>> s = m.write_ntriples_string(format="turtle")

        :param format: One of "ntriples", "turtle", "rdf/xml".
        :param graph: The IRI of the graph to write.
        :return Triples in model in the NTriples format (potentially a large string)
        """

    def write_native_parquet(
        self, folder_path: Union[str, Path], graph: str = None
    ) -> None:
        """
        Write non-transient triples using the internal native Parquet format.

        Usage:

        >>> m.write_native_parquet("output_folder")

        :param folder_path: The path of the folder to write triples in the native format.
        :param graph: The IRI of the graph to write.
        """

    def detach_graph(self, graph:str=None, preserve_name:bool=False) -> "Model":
        """
        Detaches and returns a named graph as their own Model object.
        The named graph is removed from the original Model.

        :param graph: The name of the graph to detach. Defaults to the default graph.
        :param preserve_name: Preserve the name of the graph in the new Model, defaults to False.
        :return: A model.
        """

    def get_predicate_iris(
        self, graph: str = None, include_transient: bool = False
    ) -> List["IRI"]:
        """
        :param graph: The graph to get the predicate iris from.
        :param include_transient: Should we include predicates only between transient triples?
        :return: The IRIs of the predicates currently in the given graph.
        """

    def get_predicate(
        self, iri: "IRI", graph: str = None, include_transient: bool = False
    ) -> List["SolutionMappings"]:
        """
        :param iri: The predicate IRI
        :param graph: The graph to get the predicate from.
        :param include_transient: Should we include transient triples?
        :return: A list of the underlying tables that store a given predicate.
        """

    def create_index(
        self, options: "IndexingOptions" = None, all: bool = True, graph: str = None
    ):
        """
        :param options: Indexing options
        :param all: Apply to all existing and new graphs
        :param graph: The graph where indexes should be added
        :return:
        """

    def infer(
        self,
        ruleset: Union[str, List[str]],
        graph: str = None,
        include_datatypes: bool = False,
        native_dataframe: bool = False,
        max_iterations: int = 100_000,
        max_results: int = 10_000_000,
        include_transient: bool = True,
        max_rows: int = 100_000_000,
        debug: bool = False,
    ) -> Optional[Dict[str, DataFrame]]:
        """
        Run the inference rules that are provided
        :param ruleset: The Datalog ruleset (a string).
        :param graph: Apply the ruleset to this graph, defaults to the default graph, or the graph specified in the rules.
        :param native_dataframe: Return columns with maplib-native formatting. Useful for round-trips.
        :param include_datatypes: Datatypes are not returned by default, set to true to return a dict with the solution mappings and the datatypes.
        :param max_iterations: Maximum number of iterations.
        :param max_results: Maximum number of results.
        :param include_transient: Include transient triples when reasoning.
        :param max_rows: Maximum estimated rows in result, helps avoid out-of-memory errors.
        :param debug: Debugs rule bodies for executions that give no triples.
        :return: The inferred N-Tuples.
        """

class MaplibException(Exception): ...

def explore(
        m: "Model",
        host: str = "localhost",
        port: int = 8000,
        bind: str = "localhost",
        popup=True,
        fts=True,
):
    """Starts a graph explorer session.
    To run from Jupyter Notebook use:
    >>> from maplib import explore
    >>>
    >>> server = explore(m)
    You can later stop the server with
    >>> server.stop()

    :param m: The Model to explore
    :param host: The hostname that we will point the browser to.
    :param port: The port where the graph explorer webserver listens on.
    :param bind: Bind to the following host / ip.
    :param popup: Pop up the browser window.
    :param fts: Enable full text search indexing
    """

def generate_templates(m: Model, graph: Optional[str]) -> Dict[str, Template]:
    """Generate templates for instantiating the classes in an ontology

    :param m: The model where the ontology is stored. We mainly rely on rdfs:subClassOf, rdfs:range and rdfs:domain.
    :param graph: The named graph where the ontology is stored.

    :return A dictionary of templates for instantiating the classes in the ontology, where the keys are the class URIs.

    Usage example - note that it is important to add the templates to the Model you want to populate."""


import polars as pl
import pytest
from datetime import datetime
from pathlib import Path
from maplib import Model, MaplibException

PATH_HERE = Path(__file__).parent
TESTDATA_PATH = PATH_HERE / "testdata"

pl.Config.set_fmt_str_lengths(300)


def test_incorrect_prefix_syntax_fails():
    """Test that incorrect :prefix syntax fails"""
    doc = """
    :prefix ex:<http://example.net/ns#>.
    ex:ExampleTemplate [?MyValue] :: {
       ottr:Triple(ex:myObject, ex:hasValue, ?MyValue)
    } .
    """
    m = Model()
    with pytest.raises(MaplibException):
        m.add_template(doc)


def test_missing_ottr_prefix_fails():
    """Test that missing ottr prefix definition fails"""
    doc = """
    @prefix ex: <http://example.net/ns#> .
    
    ex:ExampleTemplate [?MyValue] :: {
       ottr:Triple(ex:myObject, ex:hasValue, ?MyValue)
    } .
    """
    m = Model()
    m.add_template(doc)


def test_all_iri_case():
    """Test IRI types in stOTTR templates"""
    doc = """
    @prefix ex: <http://example.net/ns#> .
    @prefix ottr: <http://ns.ottr.xyz/0.4/> .
    
    ex:ExampleTemplate [ottr:IRI ?myVar1] :: {
        ottr:Triple(ex:anObject, ex:relatesTo, ?myVar1)
    } .
    """
    m = Model()
    m.add_template(doc)

    df = pl.DataFrame(
        {
            "myVar1": [
                "http://example.net/ns#OneThing",
                "http://example.net/ns#AnotherThing",
            ]
        }
    )

    m.map("ex:ExampleTemplate", df)

    result = m.query(
        """
        PREFIX ex: <http://example.net/ns#>
        SELECT ?o WHERE {
            ex:anObject ex:relatesTo ?o .
        }
        ORDER BY ?o
    """
    )

    assert result.height == 2
    assert result["o"][0] == "<http://example.net/ns#AnotherThing>"
    assert result["o"][1] == "<http://example.net/ns#OneThing>"


def test_string_language_tag_cases():
    """Test strings and language tags"""
    doc = """
    @prefix ex: <http://example.net/ns#> .
    @prefix ottr: <http://ns.ottr.xyz/0.4/> .
    
    ex:ExampleTemplate [?myString] :: {
        ottr:Triple(ex:anObject, ex:hasString, ?myString),
        ottr:Triple(ex:anotherObject, ex:hasString, ""@ar-SA)
    } .
    """
    m = Model()
    m.add_template(doc)

    df = pl.DataFrame({"myString": ["one", "two"]})

    m.map("ex:ExampleTemplate", df)

    result = m.query(
        """
        PREFIX ex: <http://example.net/ns#>
        SELECT ?s ?o WHERE {
            ?s ex:hasString ?o .
        }
    """
    )

    # Should have 2 rows from myString + 1 row from empty language tagged literal
    assert result.height == 3


def test_const_list_case():
    """Test constant list expansion"""
    doc = """
    @prefix ex: <http://example.net/ns#> .
    @prefix ottr: <http://ns.ottr.xyz/0.4/> .
    
    ex:ExampleTemplate [ottr:IRI ?var1] :: {
        cross | ottr:Triple(?var1, ex:hasNumber, ++(1, 2))
    } .
    """
    m = Model()
    m.add_template(doc)

    df = pl.DataFrame(
        {
            "var1": [
                "http://example.net/ns#OneThing",
                "http://example.net/ns#AnotherThing",
            ]
        }
    )

    m.map("ex:ExampleTemplate", df)

    result = m.query(
        """
        PREFIX ex: <http://example.net/ns#>
        SELECT ?s ?o WHERE {
            ?s ex:hasNumber ?o .
        }
        ORDER BY ?s ?o
    """
    )

    # 2 subjects × 2 numbers = 4 triples
    assert result.height == 4


def test_nested_templates():
    """Test template calling another template"""
    doc = """
    @prefix ex: <http://example.net/ns#> .
    @prefix ottr: <http://ns.ottr.xyz/0.4/> .
    
    ex:ExampleTemplate [?myVar1, ?myVar2] :: {
        ex:Nested(?myVar1),
        ottr:Triple(ex:anObject, ex:hasOtherNumber, ?myVar2)
    } .
    
    ex:Nested [?myVar] :: {
        ottr:Triple(ex:anObject, ex:hasNumber, ?myVar)
    } .
    """
    m = Model()
    m.add_template(doc)

    df = pl.DataFrame({"myVar1": [1, 2], "myVar2": [3, 4]})

    m.map("ex:ExampleTemplate", df)

    result = m.query(
        """
        PREFIX ex: <http://example.net/ns#>
        SELECT ?p ?o WHERE {
            ex:anObject ?p ?o .
        }
        ORDER BY ?p ?o
    """
    )

    # Should have 4 triples: 2 hasNumber + 2 hasOtherNumber
    assert result.height == 4


def test_derived_datatypes():
    """Test various Polars datatypes being mapped to RDF"""
    doc = """
    @prefix ex: <http://example.net/ns#> .
    @prefix ottr: <http://ns.ottr.xyz/0.4/> .
    
    ex:ExampleTemplate [
        ?Boolean,
        ?UInt32,
        ?UInt64,
        ?Int32,
        ?Int64,
        ?Float32,
        ?Float64,
        ?String,
        ?Datetime
    ] :: {
        ottr:Triple(ex:anObject, ex:hasVal, ?Boolean),
        ottr:Triple(ex:anObject, ex:hasVal, ?UInt32),
        ottr:Triple(ex:anObject, ex:hasVal, ?UInt64),
        ottr:Triple(ex:anObject, ex:hasVal, ?Int32),
        ottr:Triple(ex:anObject, ex:hasVal, ?Int64),
        ottr:Triple(ex:anotherObject, ex:hasValVal, ?Float32),
        ottr:Triple(ex:anotherObject, ex:hasValVal, ?Float64),
        ottr:Triple(ex:yetAnotherObject, ex:hasString, ?String),
        ottr:Triple(ex:yetAnotherObject, ex:hasDateTime, ?Datetime)
    } .
    """
    m = Model()
    m.add_template(doc)

    df = pl.DataFrame(
        {
            "Boolean": [True, False],
            "UInt32": pl.Series([5, 6], dtype=pl.UInt32),
            "UInt64": pl.Series([7, 8], dtype=pl.UInt64),
            "Int32": pl.Series([-13, -14], dtype=pl.Int32),
            "Int64": pl.Series([-15, -16], dtype=pl.Int64),
            "Float32": pl.Series([17.18, 19.20], dtype=pl.Float32),
            "Float64": [21.22, 23.24],
            "String": ["abcde", "fghij"],
            "Datetime": [
                datetime(2022, 7, 3, 10, 6, 20, 123000),
                datetime(2022, 7, 3, 10, 6, 21, 456000),
            ],
        }
    )

    m.map("ex:ExampleTemplate", df)

    result = m.query(
        """
        PREFIX ex: <http://example.net/ns#>
        SELECT ?s ?p ?o WHERE {
            ?s ?p ?o .
        }
    """
    )

    # Each row maps to 9 triples = 18 triples total
    assert result.height == 18


def test_list_arguments():
    """Test list expansion with grouping"""
    doc = """
    @prefix ex: <http://example.net/ns#> .
    @prefix ottr: <http://ns.ottr.xyz/0.4/> .
    
    ex:AnotherExampleTemplate [ottr:IRI ?object, ottr:IRI ?predicate, ?myList] :: {
        cross | ottr:Triple(?object, ?predicate, ++?myList)
    } .
    """
    m = Model()
    m.add_template(doc)

    # Create data with groups
    df = pl.DataFrame(
        {
            "object": [
                "http://example.net/ns#obj1",
                "http://example.net/ns#obj1",
                "http://example.net/ns#obj2",
                "http://example.net/ns#obj2",
            ],
            "predicate": [
                "http://example.net/ns#hasNumberFromList1",
                "http://example.net/ns#hasNumberFromList1",
                "http://example.net/ns#hasNumberFromList2",
                "http://example.net/ns#hasNumberFromList2",
            ],
            "myList": [1, 2, 3, 4],
        }
    )

    # Group by object and predicate
    df = df.group_by(["object", "predicate"], maintain_order=True).agg(pl.col("myList"))

    m.map("ex:AnotherExampleTemplate", df)

    result = m.query(
        """
        PREFIX ex: <http://example.net/ns#>
        SELECT ?s ?p ?o WHERE {
            ?s ?p ?o .
        }
        ORDER BY ?s ?p ?o
    """
    )

    # obj1 with 2 values, obj2 with 2 values = 4 triples
    assert result.height == 4


def test_two_list_arguments():
    """Test expansion of two lists"""
    doc = """
    @prefix ex: <http://example.net/ns#> .
    @prefix ottr: <http://ns.ottr.xyz/0.4/> .
    
    ex:TwoListTemplate [ottr:IRI ?object, ?myList1, ?myList2] :: {
        cross | ottr:Triple(?object, ex:hasNumber, ++?myList1),
        cross | ottr:Triple(?object, ex:hasOtherNumber, ++?myList2)
    } .
    """
    m = Model()
    m.add_template(doc)

    df = pl.DataFrame(
        {
            "object": ["http://example.net/ns#obj1", "http://example.net/ns#obj2"],
            "myList1": [[1, 2], [3, 4]],
            "myList2": [[5, 6], [7, 8, 9]],
        }
    )

    m.map("ex:TwoListTemplate", df)

    result = m.query(
        """
        PREFIX ex: <http://example.net/ns#>
        SELECT ?s ?p ?o WHERE {
            ?s ?p ?o .
        }
        ORDER BY ?s ?p ?o
    """
    )

    # obj1: 2 hasNumber + 2 hasOtherNumber = 4
    # obj2: 2 hasNumber + 3 hasOtherNumber = 5
    # Total = 9 triples
    assert result.height == 9


def test_default():
    """Test default model without explicit template"""
    m = Model()

    df = pl.DataFrame(
        {
            "subject": [
                "http://example.net/ns#obj1",
                "http://example.net/ns#obj1",
                "http://example.net/ns#obj2",
            ],
            "myVar1": [1, 2, None],
            "myVar2": [5, 6, 7],
        }
    )

    m.map_default(df, "subject")

    result = m.query(
        """
        PREFIX ex: <http://example.net/ns#>
        SELECT ?s ?p ?o WHERE {
            ?s ?p ?o .
        }
        ORDER BY ?s ?p ?o
    """
    )

    # obj1: 2 myVar1 values + 2 myVar2 values = 4 triples
    # obj2: 0 myVar1 values (null) + 1 myVar2 value = 1 triple
    # Total = 5 triples
    assert result.height == 5


def test_default_list():
    """Test default model with lists"""
    m = Model()

    df = pl.DataFrame(
        {
            "subject": [
                "http://example.net/ns#obj1",
                "http://example.net/ns#obj1",
                "http://example.net/ns#obj2",
            ],
            "myList1": [1, 2, None],
            "myList2": [5, 6, 7],
        }
    )

    # Group by subject and aggregate into lists
    df = df.group_by("subject", maintain_order=True).agg(
        [pl.col("myList1").drop_nulls(), pl.col("myList2")]
    )

    m.map_default(df, "subject")

    result = m.query(
        """
        PREFIX ex: <http://example.net/ns#>
        SELECT ?s ?p ?o WHERE {
            ?s ?p ?o .
        }
        ORDER BY ?s ?p ?o
    """
    )

    # obj1: [1, 2] for myList1 + [5, 6] for myList2 = 4 triples
    # obj2: [] for myList1 (nulls dropped) + [7] for myList2 = 1 triple
    # Total = 5 triples
    assert result.height == 5


def test_comments_in_template():
    """Test that comments are properly ignored"""
    doc = """
    @prefix ex: <http://example.net/ns#> .
    @prefix ottr: <http://ns.ottr.xyz/0.4/> .
    
    # This is a comment at the start
    ex:ExampleTemplate [?value] :: {
        # Comment before triple
        ottr:Triple(ex:subject, ex:property, ?value)  # inline comment
        # Comment after triple
    } .
    """
    m = Model()
    m.add_template(doc)

    df = pl.DataFrame({"value": ["test"]})
    m.map("ex:ExampleTemplate", df)

    result = m.query("SELECT ?o WHERE { ?s ?p ?o }")
    assert result.height == 1


def test_optional_parameters():
    """Test optional parameters with ? prefix"""
    doc = """
    @prefix ex: <http://example.net/ns#> .
    @prefix ottr: <http://ns.ottr.xyz/0.4/> .
    @prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
    
    ex:OptionalTemplate [xsd:string ?required, ? xsd:string ?optional] :: {
        ottr:Triple(ex:subject, ex:hasRequired, ?required),
        ottr:Triple(ex:subject, ex:hasOptional, ?optional)
    } .
    """
    m = Model()
    m.add_template(doc)

    # Test with both values present
    df1 = pl.DataFrame({"required": ["value1"], "optional": ["value2"]})
    m.map("ex:OptionalTemplate", df1)

    # Test with optional missing (null)
    df2 = pl.DataFrame({"required": ["value3"], "optional": [None]})
    m.map("ex:OptionalTemplate", df2)

    result = m.query(
        """
        PREFIX ex: <http://example.net/ns#>
        SELECT ?p ?o WHERE {
            ex:subject ?p ?o .
        }
        ORDER BY ?p ?o
    """
    )

    # Should have: 2 required + 1 optional (null is skipped) = 3 triples
    assert result.height == 3


def test_typed_literals():
    """Test literals with explicit type annotations"""
    doc = """
    @prefix ex: <http://example.net/ns#> .
    @prefix ottr: <http://ns.ottr.xyz/0.4/> .
    @prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
    
    ex:TypedLiteralTemplate [xsd:string ?id] :: {
        ottr:Triple(ex:subject, ex:country, "United Kingdom"^^xsd:string),
        ottr:Triple(ex:subject, ex:hasId, ?id)
    } .
    """
    m = Model()
    m.add_template(doc)

    df = pl.DataFrame({"id": ["test"]})
    m.map("ex:TypedLiteralTemplate", df)

    result = m.query(
        """
        PREFIX ex: <http://example.net/ns#>
        SELECT ?o WHERE {
            ex:subject ex:country ?o .
        }
    """
    )

    assert result.height == 1
    assert result["o"][0] == "United Kingdom"


def test_multiline_parameters():
    """Test parameter lists split across multiple lines"""
    doc = """
    @prefix ex: <http://example.net/ns#> .
    @prefix ottr: <http://ns.ottr.xyz/0.4/> .
    @prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
    
    ex:MultilineParamsTemplate [
        xsd:string ?param1,
        xsd:string ?param2,
        ? xsd:string ?param3
    ] :: {
        ottr:Triple(ex:subject, ex:prop1, ?param1),
        ottr:Triple(ex:subject, ex:prop2, ?param2),
        ottr:Triple(ex:subject, ex:prop3, ?param3)
    } .
    """
    m = Model()
    m.add_template(doc)

    df = pl.DataFrame({"param1": ["a"], "param2": ["b"], "param3": ["c"]})
    m.map("ex:MultilineParamsTemplate", df)

    result = m.query(
        """
        PREFIX ex: <http://example.net/ns#>
        SELECT ?p ?o WHERE {
            ex:subject ?p ?o .
        }
    """
    )

    assert result.height == 3


def test_multiline_instances():
    """Test instance lists split across multiple lines"""
    doc = """
    @prefix ex: <http://example.net/ns#> .
    @prefix ottr: <http://ns.ottr.xyz/0.4/> .
    
    ex:MultilineInstancesTemplate [?val] :: {
        ottr:Triple(
            ex:subject,
            ex:property,
            ?val
        ),
        ottr:Triple(
            ex:subject,
            ex:anotherProperty,
            ?val
        )
    } .
    """
    m = Model()
    m.add_template(doc)

    df = pl.DataFrame({"val": ["test"]})
    m.map("ex:MultilineInstancesTemplate", df)

    result = m.query(
        """
        PREFIX ex: <http://example.net/ns#>
        SELECT ?p WHERE {
            ex:subject ?p ?o .
        }
        ORDER BY ?p
    """
    )

    assert result.height == 2


def test_xsd_anyuri_type():
    """Test xsd:anyURI parameter type"""
    doc = """
    @prefix ex: <http://example.net/ns#> .
    @prefix ottr: <http://ns.ottr.xyz/0.4/> .
    @prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
    
    ex:URITemplate [? xsd:anyURI ?website] :: {
        ottr:Triple(ex:subject, ex:hasWebsite, ?website)
    } .
    """
    m = Model()
    m.add_template(doc)

    df = pl.DataFrame({"website": ["https://example.com", None]})
    m.map("ex:URITemplate", df)

    result = m.query(
        """
        PREFIX ex: <http://example.net/ns#>
        SELECT ?o WHERE {
            ex:subject ex:hasWebsite ?o .
        }
    """
    )

    # Only non-null value should appear
    assert result.height == 1
    assert "example.com" in result["o"][0]


def test_long_iri_prefixes():
    """Test templates with full IRI names instead of prefix abbreviations"""
    doc = """
    @prefix ottr: <http://ns.ottr.xyz/0.4/> .
    @prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
    
    <https://example.org/very/long/path/to/template#MyTemplate> [xsd:string ?value] :: {
        ottr:Triple(
            <https://example.org/very/long/path/to/entity#test>,
            <https://example.org/very/long/path/to/property#hasValue>,
            ?value
        )
    } .
    """
    m = Model()
    m.add_template(doc)

    df = pl.DataFrame({"value": ["test"]})
    m.map("https://example.org/very/long/path/to/template#MyTemplate", df)

    result = m.query(
        """
        SELECT ?s WHERE {
            ?s <https://example.org/very/long/path/to/property#hasValue> "test" .
        }
    """
    )

    assert result.height == 1


def test_easy_template_bad_prefix():
    doc = """
    @prefix ex: <http://example.org#>.
        ex:Person[ ?firstName, ?lastName, ?email ] :: {
            o-rdf:Type (_:person, foaf:Person )
        } .
    """
    m = Model()
    with pytest.raises(MaplibException):
        m.add_template(doc)


def test_easy_template_extra_comma():
    doc = """
    @prefix ex: <http://example.org#>.
    @prefix foaf: <http://xmlns.com/foaf/0.1/>.
    ex:Person[ ?firstName, ?lastName, ?email, ] :: {
        ex:Type (_:person, foaf:Person )
    } .
    """
    m = Model()
    m.add_template(doc)


def test_easy_template_sparql_prefix():
    doc = """
    prefix ex: <http://example.org#>
    @prefix foaf: <http://xmlns.com/foaf/0.1/>.
    ex:Person[ ?firstName, ?lastName, ?email, ] :: {
        ex:Type (_:person, foaf:Person )
    } .
    """
    m = Model()
    m.add_template(doc)


def test_easy_template_sparql_prefix2():
    doc = """
    @prefix ex: <http://example.org#> .
    prefix foaf: <http://xmlns.com/foaf/0.1/>
    ex:Person[ ?firstName, ?lastName, ?email, ] :: {
        ex:Type (_:person, foaf:Person )
    } .
    """
    m = Model()
    m.add_template(doc)


def test_spec_modifiers_1():
    doc = """
    @prefix ex:<http://example.net/ns#>.
    ex:NamedPizza [ ??pizza  ] .
    """
    m = Model()
    m.add_template(doc)


def test_spec_modifiers_2():
    doc = """
    @prefix ex:<http://example.net/ns#>.

    ex:NamedPizza [ !?pizza  ] .
    """
    m = Model()
    m.add_template(doc)


def test_spec_modifiers_3():
    doc = """
    @prefix ex:<http://example.net/ns#>.

    ex:NamedPizza [ ?!?pizza  ] .
    """
    m = Model()
    m.add_template(doc)


def test_spec_modifiers_4():
    doc = """
    @prefix ex:<http://example.net/ns#>.

    ex:NamedPizza [ !??pizza  ] .
    """
    m = Model()
    m.add_template(doc)


def test_spec_type_1():
    doc = """
    @prefix ex:<http://example.net/ns#>.
    ex:NamedPizza [ owl:Class ?pizza ].
    """
    m = Model()
    m.add_template(doc)


def test_spec_type_2():
    doc = """
    @prefix ex:<http://example.net/ns#>.

    ex:NamedPizza [ ? owl:Class ?pizza ].
    """
    m = Model()
    m.add_template(doc)


def test_spec_type_3():
    doc = """
    @prefix ex:<http://example.net/ns#>.

    ex:NamedPizza [ ?! owl:Class ?pizza ].
    """
    m = Model()
    m.add_template(doc)


def test_spec_default_value_1():
    doc = """
    @prefix ex:<http://example.net/ns#>.
    @prefix p:<http://example.net/pizzas#>.
    ex:NamedPizza[ owl:Class ?pizza = p:pizza] .
    """
    m = Model()
    m.add_template(doc)


def test_spec_default_value_2():
    doc = """
    @prefix ex:<http://example.net/ns#>.
    @prefix p:<http://example.net/pizzas#>.
    ex:NamedPizza[ owl:Class ?pizza = 2] .
    """
    m = Model()
    m.add_template(doc)


def test_spec_default_value_3():
    doc = """
    @prefix ex:<http://example.net/ns#>.
    @prefix p:<http://example.net/pizzas#>.
    ex:NamedPizza[ owl:Class ?pizza = "asdf"] .
    """
    m = Model()
    m.add_template(doc)


def test_spec_more_parameters():
    doc = """
    @prefix ex:<http://example.net/ns#>.
    @prefix p:<http://example.net/pizzas#>.
    ex:NamedPizza [  ?pizza ,  ?country  ,  ?toppings ] .
    """
    m = Model()
    m.add_template(doc)


def test_spec_lists():
    doc = """
    @prefix ex:<http://example.net/ns#>.
    @prefix p:<http://example.net/pizzas#>.
    ex:NamedPizza [  ?pizza = "asdf" ,  ?country = ("asdf", "asdf") ,  ?toppings = ((())) ] .
    """
    m = Model()
    m.add_template(doc)


def test_spec_more_complex_types():
    doc = """
    @prefix ex:<http://example.net/ns#>.
    @prefix p:<http://example.net/pizzas#>.
    ex:NamedPizza [
      ! owl:Class ?pizza  ,
      ?! owl:NamedIndividual ?country  = ex:Class ,
      NEList<List<List<owl:Class>>> ?toppings
    ] ."""
    m = Model()
    m.add_template(doc)


def test_missing_dot():
    doc = """
    @prefix ex:<http://example.net/ns#>.
    @prefix p:<http://example.net/pizzas#>.
    ex:template [ ] :: { ex:template((ex:template)) } 
    """
    m = Model()
    m.add_template(doc)


def test_spec_example_1():
    doc = """
    @prefix ex:<http://example.net/ns#>.
    @prefix p:<http://example.net/pizzas#>.
    ex:template [ ] :: { ex:template((ex:template)) } .
    """
    m = Model()
    m.add_template(doc)


def test_spec_example_2():
    doc = """
    @prefix ex:<http://example.net/ns#>.
    @prefix p:<http://example.net/pizzas#>.
    ex:template [?!?var ] :: { ex:template((((ex:template)))) } .
    """
    m = Model()
    m.add_template(doc)


def test_spec_example_3():
    doc = """
    @prefix ex:<http://example.net/ns#>.
    @prefix p:<http://example.net/pizzas#>.
    ex:template [ ] :: { ex:template(( ex:template )) } .
    """
    m = Model()
    m.add_template(doc)


def test_spec_example_4():
    doc = """
    @prefix ex:<http://example.net/ns#>.
    @prefix p:<http://example.net/pizzas#>.
    ex:NamedPizza [
      ! owl:Class ?pizza = p:Grandiosa , ?! LUB<owl:NamedIndividual> ?country  , List<owl:Class> ?toppings
      ]
      @@ cross | ex:SomeAnnotationTemplate("asdf", "asdf", "asdf" ),
      @@<http://asdf>("asdf", "asdf", ++("A", "B", "C") )
      :: {
         cross | ex:Template1 (?pizza, ++?toppings) ,
         ex:Template2 (1, 2,4,   5) ,
         <http://Template2.com> ("asdf"^^xsd:string) ,
         zipMax | ex:Template4 ("asdf"^^xsd:string, ?pizza, ++( "a", "B" )),
         zipMax | ex:Template4 ([], [], [], ++([], []))
      } .
    """
    m = Model()
    m.add_template(doc)


def test_trailing_commas_and_rdftype_shorthand():
    doc = """
    @prefix rr:<http://www.w3.org/ns/r2rml#>.
@prefix foaf:<http://xmlns.com/foaf/0.1/>.
@prefix xsd:<http://www.w3.org/2001/XMLSchema#>.
@prefix rdfs:<http://www.w3.org/2000/01/rdf-schema#>.
@prefix dc:<http://purl.org/dc/elements/1.1/>.
@prefix rev:<http://purl.org/stuff/rev#>.
@prefix gtfs:<http://vocab.gtfs.org/terms#>.
@prefix geo:<http://www.w3.org/2003/01/geo/wgs84_pos#>.
@prefix schema:<http://schema.org/>.
@prefix dct:<http://purl.org/dc/terms/>.
@prefix rml:<http://semweb.mmlab.be/ns/rml#>.
@prefix ql:<http://semweb.mmlab.be/ns/ql#>.
@prefix rdf:<http://www.w3.org/1999/02/22-rdf-syntax-ns#>.
@prefix t:<https://github.com/magbak/maplib/benchmark/template#>.

t:Stops [ ottr:IRI ?stop_id, ?stop_code, ?stop_name, ?stop_desc, ?stop_lat, ?stop_lon, ??zone_id,
          ottr:IRI ?stop_url, ottr:IRI ?location_type, ? ottr:IRI ?parent_station, ??stop_timezone, ?wheelchair_boarding
          , ] :: {
  ottr:Triple(?stop_id, a, gtfs:Stop) ,
  ottr:Triple(?stop_id, gtfs:code, ?stop_code) ,
  ottr:Triple(?stop_id, dct:identifier, ?stop_id) ,
  ottr:Triple(?stop_id, foaf:name, ?stop_name) ,
  ottr:Triple(?stop_id, dct:description, ?stop_desc) ,
  ottr:Triple(?stop_id, geo:lat, ?stop_lat) ,
  ottr:Triple(?stop_id, geo:long, ?stop_lon) ,
  ottr:Triple(?stop_id, gtfs:zone, ?zone_id) ,
  ottr:Triple(?stop_id, foaf:page, ?stop_url) ,
  ottr:Triple(?stop_id, gtfs:locationType, ?location_type) ,
  ottr:Triple(?stop_id, gtfs:parentStation, ?parent_station) ,
  ottr:Triple(?stop_id, gtfs:timeZone, ?stop_timezone) ,
  ottr:Triple(?stop_id, gtfs:wheelchairAccessible, ?wheelchair_boarding)

  ,
} . 
    """
    m = Model()
    m.add_template(doc)


def test_handles_space_before_prefix_dot():
    doc = """
    @prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
    @prefix foaf: <http://xmlns.com/foaf/0.1/>.
    @prefix ex: <http://example.org#>.

    ex:Person[ ?firstName, ?lastName, ?email ] :: {
      ottr:Triple(_:person, rdf:type, foaf:Person ),
      ottr:Triple(_:person, foaf:firstName, ?firstName ),
      ottr:Triple(_:person, foaf:lastName, ?lastName ),
      ottr:Triple(_:person, foaf:mbox, ?email )
    } .
    """
    m = Model()
    m.add_template(doc)


def test_optional_commas():
    doc = """
    @prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
    @prefix ex: <http://example.org#>.
    @prefix foaf: <http://xmlns.com/foaf/0.1/>.

    ex:Person[ ?firstName, ?lastName, ?email ] :: {
      ottr:Triple(_:person, rdf:type, foaf:Person )
      ottr:Triple(_:person, foaf:firstName, ?firstName ),
      ottr:Triple(_:person, foaf:lastName, ?lastName )
      ottr:Triple(_:person, foaf:mbox, ?email )
    } .
    """
    m = Model()
    m.add_template(doc)

def test_read_template():
    m = Model()
    m.read_template(TESTDATA_PATH / "templates.ttl")
    df1 = pl.DataFrame(
        {
            "myVar1": [
                "http://example.net/ns#OneThing1",
                "http://example.net/ns#AnotherThing1",
            ]
        }
    )
    df2 = pl.DataFrame(
        {
            "myVar1": [
                "http://example.net/ns#OneThing2",
                "http://example.net/ns#AnotherThing2",
            ]
        }
    )

    m.map("ex:ExampleTemplate1", df1)
    m.map("http://example.net/ns#ExampleTemplate2", df2)

    df = m.query("""SELECT * WHERE {?a ?b ?c}""")
    assert df.height == 4



from maplib.maplib import Model, Template, IRI, Triple, Variable


def add_triples(
    source: Model, target: Model, source_graph: str = None, target_graph: str = None
):
    """(Zero) copy the triples from one Model into another.

    :param source: The source model
    :param target: The target model
    :param source_graph: The named graph in the source model to copy from. None means default graph.
    :param target_graph: The named graph in the target model to copy into. None means default graph.
    """
    for p in source.get_predicate_iris(source_graph):
        subject = Variable("subject")
        object = Variable("object")
        template = Template(
            iri=IRI("urn:maplib:tmp"),
            parameters=[subject, object],
            instances=[Triple(subject, p, object)],
        )
        sms = source.get_predicate(p, source_graph)
        for sm in sms:
            target.map(
                template,
                sm.mappings,
                types=sm.rdf_types,
                graph=target_graph,
            )



import pytest
from maplib import Model
from maplib import MaplibException


def test_model_exception():
    m = Model()

    with pytest.raises(MaplibException) as e:
        m.reads("abc", format="turtle", graph="http://example.com/data")



import polars as pl
import pytest
import rdflib
from polars.testing import assert_frame_equal
import pathlib
from maplib import Model

pl.Config.set_fmt_str_lengths(300)


PATH_HERE = pathlib.Path(__file__).parent
TESTDATA_PATH = PATH_HERE / "testdata"


def test_read_ntriples():
    m = Model()
    m.read(str(TESTDATA_PATH / "read_ntriples.nt"))
    res = m.query(
        """
            PREFIX foaf:<http://xmlns.com/foaf/0.1/>

            SELECT ?s ?v ?o WHERE {
            ?s ?v ?o .
            } ORDER BY ?s ?v ?o
            """
    ).sort(["s", "v", "o"])
    # TODO: Fix multitype sorting
    filename = TESTDATA_PATH / "read_ntriples.csv"
    # res.write_csv(str(filename))
    expected_df = pl.scan_csv(filename).collect()
    pl.testing.assert_frame_equal(res, expected_df)


def test_read_ntriples_twice_with_replace():
    m = Model()
    m.read(str(TESTDATA_PATH / "read_ntriples.nt"))
    m.read(str(TESTDATA_PATH / "read_ntriples.nt"), replace_graph=True)
    res = m.query(
        """
            PREFIX foaf:<http://xmlns.com/foaf/0.1/>

            SELECT ?s ?v ?o WHERE {
            ?s ?v ?o .
            } ORDER BY ?s ?v ?o
            """
    ).sort(["s", "v", "o"])
    # TODO: Fix multitype sorting
    filename = TESTDATA_PATH / "read_ntriples.csv"
    # res.write_csv(str(filename))
    assert res.shape == (8, 3)


def test_read_write_ntriples_string():
    m = Model()
    with open(TESTDATA_PATH / "read_ntriples.nt") as f:
        ntstring = f.read()
    m.reads(ntstring, format="ntriples")
    out_str = m.writes(format="ntriples")
    m2 = Model()
    m2.reads(out_str, format="ntriples")
    res = m2.query(
        """
            PREFIX foaf:<http://xmlns.com/foaf/0.1/>

            SELECT ?v ?o WHERE {
            ?s ?v ?o .
            } ORDER BY ?v ?o
            """
    ).sort(["v", "o"])
    # TODO: Fix multitype sorting
    filename = TESTDATA_PATH / "read_ntriples2.csv"
    # res.write_csv(str(filename))
    expected_df = pl.scan_csv(filename).select(["v", "o"]).sort(["v", "o"]).collect()
    pl.testing.assert_frame_equal(res, expected_df)


def test_read_write_turtle_string():
    m = Model()
    with open(TESTDATA_PATH / "read_ntriples.nt") as f:
        ntstring = f.read()
    m.reads(ntstring, format="ntriples")
    out_str = m.writes(format="turtle")
    m2 = Model()
    m2.reads(out_str, format="turtle")
    res = m2.query(
        """
            PREFIX foaf:<http://xmlns.com/foaf/0.1/>

            SELECT ?v ?o WHERE {
            ?s ?v ?o .
            } ORDER BY ?v ?o
            """
    ).sort(["v", "o"])
    # TODO: Fix multitype sorting
    filename = TESTDATA_PATH / "read_ntriples2.csv"
    # res.write_csv(str(filename))
    expected_df = pl.scan_csv(filename).select(["v", "o"]).sort(["v", "o"]).collect()
    pl.testing.assert_frame_equal(res, expected_df)


def test_read_write_xml_string():
    m = Model()
    with open(TESTDATA_PATH / "read_ntriples.nt") as f:
        ntstring = f.read()
    m.reads(ntstring, format="ntriples")
    out_str = m.writes(format="rdf/xml")
    m2 = Model()
    m2.reads(out_str, format="rdf/xml")
    res = m2.query(
        """
            PREFIX foaf:<http://xmlns.com/foaf/0.1/>

            SELECT ?v ?o WHERE {
            ?s ?v ?o .
            } ORDER BY ?v ?o
            """
    ).sort(["v", "o"])
    # TODO: Fix multitype sorting
    filename = TESTDATA_PATH / "read_ntriples2.csv"
    # res.write_csv(str(filename))
    expected_df = pl.scan_csv(filename).select(["v", "o"]).sort(["v", "o"]).collect()
    pl.testing.assert_frame_equal(res, expected_df)



import polars as pl
from typing import List


def topological_sort(subclasses: pl.DataFrame) -> List[str]:
    in_degree_df = subclasses.group_by("child").agg([pl.col("parent").count()])
    in_degree_dict = {}
    for d in in_degree_df.iter_rows(named=True):
        in_degree_dict[d["child"]] = d["parent"]

    out_edges_df = subclasses.group_by("parent").agg([pl.col("child")])
    out_edges_dict = {}
    for d in out_edges_df.iter_rows(named=True):
        out_edges_dict[d["parent"]] = d["child"]

    sorting = []

    # First, we can add those parents who have no parents.
    for p in subclasses.select("parent").unique().iter_rows(named=True):
        if p["parent"] not in in_degree_dict:
            sorting.append(p["parent"])
            for out_edge in out_edges_dict[p["parent"]]:
                in_degree_dict[out_edge] = in_degree_dict[out_edge] - 1

    # Next, a topological sort defines ordering.
    while len(in_degree_dict) > 0:
        found_p = None
        for p, i in in_degree_dict.items():
            if i == 0:
                found_p = p
                break
        if found_p is None:
            assert False, "Error in the algorithm.. "

        sorting.append(found_p)
        in_degree_dict.pop(found_p)
        if found_p in out_edges_dict:
            for out_edge in out_edges_dict[found_p]:
                in_degree_dict[out_edge] = in_degree_dict[out_edge] - 1
    return sorting



import polars as pl
import pytest
from polars.testing import assert_frame_equal

from maplib import Model, RDFType

pl.Config.set_fmt_str_lengths(300)


@pytest.fixture(scope="function")
def pizzas_model():
    doc = """
    @prefix pizza:<https://github.com/magbak/maplib/pizza#>.
    @prefix xsd:<http://www.w3.org/2001/XMLSchema#>.
    @prefix ex:<https://github.com/magbak/maplib/pizza#>.

    ex:Pizza[?p, xsd:anyURI ?c, List<ottr:IRI> ?is] :: {
    ottr:Triple(?p, a, pizza:Pizza),
    ottr:Triple(?p, pizza:fromCountry, ?c),
    cross | ottr:Triple(?p, pizza:hasIngredient, ++?is)
    }.
    """

    m = Model()
    m.add_template(doc)

    co = "https://github.com/magbak/maplib/countries#"
    pi = "https://github.com/magbak/maplib/pizza#"
    ing = "https://github.com/magbak/maplib/pizza/ingredients#"

    df = pl.DataFrame(
        {
            "p": [pi + "Hawaiian", pi + "Grandiosa"],
            "c": [co + "CAN", co + "NOR"],
            "is": [[ing + "Pineapple", ing + "Ham"], [ing + "Pepper", ing + "Meat"]],
        }
    )
    m.map("ex:Pizza", df)
    hpizzas = """
    PREFIX pizza:<https://github.com/magbak/maplib/pizza#>
    PREFIX ing:<https://github.com/magbak/maplib/pizza/ingredients#>
    CONSTRUCT { ?p a pizza:HeterodoxPizza } 
    WHERE {
        ?p a pizza:Pizza .
        ?p pizza:hasIngredient ing:Pineapple .
    }"""
    m.insert(hpizzas)
    return m


def test_simple_query_no_error(pizzas_model):
    res = pizzas_model.query(
        """
    PREFIX pizza:<https://github.com/magbak/maplib/pizza#>

    SELECT ?p WHERE {
    ?p a pizza:HeterodoxPizza
    }
    """
    )

    expected_df = pl.DataFrame(
        {"p": ["<https://github.com/magbak/maplib/pizza#Hawaiian>"]}
    )
    assert_frame_equal(res, expected_df)


def test_construct_pvalues(pizzas_model):
    h_df = pl.DataFrame(
        {
            "h1": [
                "https://github.com/magbak/maplib/pizza#Hawaiian1",
                "https://github.com/magbak/maplib/pizza#Hawaiian2",
            ],
            "h2": [
                "https://github.com/magbak/maplib/pizza#Hawaiian3",
                "https://github.com/magbak/maplib/pizza#Hawaiian4",
            ],
        }
    )

    res = pizzas_model.query(
        """
    PREFIX pizza:<https://github.com/magbak/maplib/pizza#>

    CONSTRUCT {
        ?h1 a pizza:Pizza . 
        ?h2 a pizza:Pizza .
    } WHERE {
        PVALUES (?h1 ?h2) h
    }
    """,
        parameters={"h": (h_df, {"h1": RDFType.IRI, "h2": RDFType.IRI})},
    )
    res0 = res[0]
    expected_dtypes = {"object": "IRI", "subject": "IRI"}
    # assert res0.rdf_datatypes == expected_dtypes
    res1 = res[1]
    # assert res1.rdf_datatypes == expected_dtypes
    assert res0.height == 2
    assert res1.height == 2


def test_construct_pvalues2(pizzas_model):
    h_df = pl.DataFrame(
        {
            "h1": [
                "https://github.com/magbak/maplib/pizza#Hawaiian",
                "https://github.com/magbak/maplib/pizza#Hawaiian2",
            ]
        }
    )
    res = pizzas_model.query(
        """
    PREFIX pizza:<https://github.com/magbak/maplib/pizza#>

    CONSTRUCT {
        ?h1 a pizza:Pizza . 
        ?i a pizza:Pizza .
    } WHERE {
        ?h1 pizza:hasIngredient ?i .
        PVALUES (?h1) h
    }
    """,
        parameters={"h": (h_df, {"h1": RDFType.IRI})},
    )
    res0 = res[0]
    res1 = res[1]
    assert res0.height == 1
    assert res1.height == 2


def test_having_not_so_nice(pizzas_model):
    res = pizzas_model.query(
        """
    PREFIX pizza:<https://github.com/magbak/maplib/pizza#>

    SELECT ?p (COUNT(?i) AS ?c)
    WHERE {
        ?p pizza:hasIngredient ?i .
    } 
    GROUP BY ?p
    HAVING (?c > 1)
    """
    )
    assert res.height == 2


def test_update_insert_delete(pizzas_model):
    pizzas_model.update(
        """
    PREFIX pizza:<https://github.com/magbak/maplib/pizza#>
    
    DELETE  {
        ?p a pizza:Pizza .
    } INSERT {
        ?p a pizza:NewPizza . 
    }WHERE {
        ?p a pizza:Pizza .
    }
    """
    )
    df = pizzas_model.query(
        """
        PREFIX pizza:<https://github.com/magbak/maplib/pizza#>
        SELECT ?a ?b ?c WHERE {
            ?a ?b ?c
        }
    """
    )
    assert df.height == 9
    df_filtered = df.filter(pl.col("c").str.contains(pl.lit("#NewPizza")))
    assert df_filtered.height == 2


def test_update_insert(pizzas_model):
    pizzas_model.update(
        """
    PREFIX pizza:<https://github.com/magbak/maplib/pizza#>
    
    INSERT {
        ?p a pizza:NewPizza . 
    }WHERE {
        ?p a pizza:Pizza .
    }
    """
    )
    df = pizzas_model.query(
        """
        PREFIX pizza:<https://github.com/magbak/maplib/pizza#>
        SELECT ?a ?b ?c WHERE {
            ?a ?b ?c
        }
    """
    )
    assert df.height == 11
    df_filtered_orig = df.filter(pl.col("c").str.contains(pl.lit("#Pizza")))
    assert df_filtered_orig.height == 2
    df_filtered_new = df.filter(pl.col("c").str.contains(pl.lit("#NewPizza")))
    assert df_filtered_new.height == 2


def test_update_delete(pizzas_model):
    pizzas_model.update(
        """
    PREFIX pizza:<https://github.com/magbak/maplib/pizza#>
    
    DELETE  {
        ?p a pizza:Pizza .
    } WHERE {
        ?p a pizza:Pizza .
    }
    """
    )
    df = pizzas_model.query(
        """
        PREFIX pizza:<https://github.com/magbak/maplib/pizza#>
        SELECT ?a ?b ?c WHERE {
            ?a ?b ?c
        }
    """
    )
    assert df.height == 7
    df_filtered_orig = df.filter(pl.col("c").str.contains(pl.lit("#Pizza")))
    assert df_filtered_orig.height == 0
    df_filtered_new = df.filter(pl.col("c").str.contains(pl.lit("#NewPizza")))
    assert df_filtered_new.height == 0


def test_update_insert_delete_multiple(pizzas_model):
    pizzas_model.update(
        """
    PREFIX pizza:<https://github.com/magbak/maplib/pizza#>
    
    DELETE  {
        ?p a pizza:Pizza .
        ?p pizza:fromCountry ?c .
    } INSERT {
        ?p a pizza:NewPizza . 
    }WHERE {
        ?p a pizza:Pizza .
        ?p pizza:fromCountry ?c .
    }
    """
    )
    df = pizzas_model.query(
        """
        PREFIX pizza:<https://github.com/magbak/maplib/pizza#>
        SELECT ?a ?b ?c WHERE {
            ?a ?b ?c
        }
    """
    )
    assert df.height == 7
    df_filtered = df.filter(pl.col("c").str.contains(pl.lit("#NewPizza")))
    assert df_filtered.height == 2
    df_filtered = df.filter(pl.col("b").str.contains(pl.lit("Country")))
    assert df_filtered.height == 0


def test_count_star(pizzas_model):
    count= pizzas_model.query( "SELECT (COUNT(*) as ?count) WHERE { ?s ?p ?o . }" )
    assert count.height == 1
    assert count.get_column("count")[0] == 9


def test_update_insert_delete_non_existant(pizzas_model):
    pizzas_model.update(
        """
    PREFIX pizza:<https://github.com/magbak/maplib/pizza#>
    
    DELETE  {
        ?s ?p ?o .
    } WHERE {
        ?s ?p ?o .
    }
    """
    )

    df = pizzas_model.query(
        """
        PREFIX pizza:<https://github.com/magbak/maplib/pizza#>
        SELECT ?a ?b ?c WHERE {
            ?a ?b ?c
        }
    """
    )
    assert df.height == 0